import abc
import dataclasses
from dataclasses import MISSING
import numpy as np
import scipy.special
import astropy.units as u
import named_arrays as na
import optika
from . import AbstractRulingSpacing

__all__ = [
    "AbstractRulings",
    "Rulings",
    "MeasuredRulings",
    "SinusoidalRulings",
    "SquareRulings",
    "SawtoothRulings",
    "TriangularRulings",
    "RectangularRulings",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRulings(
    optika.mixins.Printable,
    optika.mixins.Shaped,
):
    """
    Interface for the interaction of a ruled surface with incident light
    """

    @property
    @abc.abstractmethod
    def diffraction_order(self) -> int | na.AbstractScalar:
        """
        the diffraction order to simulate
        """

    @property
    @abc.abstractmethod
    def spacing(
        self,
    ) -> u.Quantity | na.AbstractScalar | AbstractRulingSpacing:
        """
        Spacing between adjacent rulings at the given position.
        """

    @property
    def spacing_(self) -> AbstractRulingSpacing:
        """
        A normalized version of :attr:`spacing` that is guaranteed to be
        an instance of :class:`optika.rulings.AbstractRulingSpacing`.
        """
        spacing = self.spacing
        if not isinstance(spacing, optika.rulings.AbstractRulingSpacing):
            spacing = optika.rulings.ConstantRulingSpacing(
                constant=spacing,
                normal=na.Cartesian3dVectorArray(1, 0, 0),
            )
        return spacing

    @abc.abstractmethod
    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> float | na.AbstractScalar:
        """
        The fraction of light that is diffracted into a given order.

        Parameters
        ----------
        rays
            The light rays incident on the rulings
        normal
            The vector normal to the surface on which the rulings are placed.
        """


@dataclasses.dataclass(eq=False, repr=False)
class Rulings(
    AbstractRulings,
):
    """
    An idealized set of rulings which have perfect efficiency in all diffraction
    orders.
    """

    spacing: u.Quantity | na.AbstractScalar | AbstractRulingSpacing = MISSING
    """Spacing between adjacent rulings at the given position."""

    diffraction_order: int | na.AbstractScalar = MISSING
    """The diffraction order to simulate."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.spacing),
            optika.shape(self.diffraction_order),
        )

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> float:
        return 1


@dataclasses.dataclass(eq=False, repr=False)
class MeasuredRulings(
    AbstractRulings,
):
    """
    A set of rulings where the efficiency has been measured or calculated
    by an independent source.
    """

    spacing: u.Quantity | na.AbstractScalar | AbstractRulingSpacing = MISSING
    """Spacing between adjacent rulings at the given position."""

    diffraction_order: int | na.AbstractScalar = MISSING
    """The diffraction order to simulate."""

    efficiency_measured: na.FunctionArray[
        na.SpectralDirectionalVectorArray,
        na.AbstractScalar,
    ] = MISSING
    """The discrete measurements of the efficiency."""

    @property
    def shape(self) -> dict[str, int]:
        axis_wavelength = self.efficiency_measured.inputs.wavelength.axes
        shape = optika.shape(self.efficiency_measured.outputs)
        for ax in axis_wavelength:
            shape.pop(ax, None)
        return na.broadcast_shapes(
            optika.shape(self.spacing),
            optika.shape(self.diffraction_order),
            shape,
        )

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:

        measurement = self.efficiency_measured

        wavelength = measurement.inputs.wavelength
        direction = measurement.inputs.direction
        efficiency = measurement.outputs

        if direction.size != 1:  # pragma: nocover
            raise ValueError(
                "Interpolating over different incidence angles is not supported."
            )

        if wavelength.ndim != 1:  # pragma: nocover
            raise ValueError(
                f"wavelength must be one dimensional, got shape {wavelength.shape}"
            )

        return na.interp(
            x=rays.wavelength,
            xp=wavelength,
            fp=efficiency,
        )


@dataclasses.dataclass(eq=False, repr=False)
class SinusoidalRulings(
    AbstractRulings,
):
    r"""
    A ruling profile described by a sinusoidal wave.

    Examples
    --------

    Compute the 1st-order groove efficiency of sinusoidal rulings with a groove
    density of 2500 grooves/mm and a groove depth of 15 nm.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the groove density
        density = 2500 / u.mm

        # Define the groove depth
        depth = 15 * u.nm

        # Define ruling model
        rulings = optika.rulings.SinusoidalRulings(
            spacing=1 / density,
            depth=depth,
            diffraction_order=1,
        )

        # Define the wavelengths at which to sample the groove efficiency
        wavelength = na.geomspace(100, 1000, axis="wavelength", num=1001) * u.AA

        # Define the incidence angles at which to sample the groove efficiency
        angle = na.linspace(0, 30, num=3, axis="angle") * u.deg

        # Define the light rays incident on the grooves
        rays = optika.rays.RayVectorArray(
            wavelength=wavelength,
            direction=na.Cartesian3dVectorArray(
                x=np.sin(angle),
                y=0,
                z=np.cos(angle),
            ),
        )

        # Compute the efficiency of the grooves for the given wavelength
        efficiency = rulings.efficiency(
            rays=rays,
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the groove efficiency as a function of wavelength
        fig, ax = plt.subplots()
        angle_str = angle.value.astype(str).astype(object)
        na.plt.plot(
            wavelength,
            efficiency,
            ax=ax,
            axis="wavelength",
            label=r"$\theta$ = " + angle_str + f"{angle.unit:latex_inline}",
        );
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel(f"efficiency");
        ax.legend();
    """

    spacing: u.Quantity | na.AbstractScalar | AbstractRulingSpacing = MISSING
    """Spacing between adjacent rulings at the given position."""

    depth: u.Quantity | na.AbstractScalar = MISSING
    """Depth of the ruling pattern."""

    diffraction_order: int | na.AbstractScalar = MISSING
    """The diffraction order to simulate."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.spacing),
            optika.shape(self.depth),
            optika.shape(self.diffraction_order),
        )

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> float | na.AbstractScalar:
        r"""
        The fraction of light diffracted into a given order.

        Calculated using the expression given in Table 1 of :cite:t:`Magnusson1978`.

        Parameters
        ----------
        rays
            The light rays incident on the rulings
        normal
            The vector normal to the surface on which the rulings are placed.

        Notes
        -----

        The theoretical efficiency of thin (wavelength much smaller than
        the groove spacing), sinusoidal rulings is given by Table 1 of
        :cite:t:`Magnusson1978`,

        .. math::

            \eta_i = J_i^2(2 \gamma)

        where :math:`\eta_i` is the groove efficiency for diffraction order
        :math:`i`, :math:`J_i(x)` is a Bessel function of the first kind,
        :math:`\gamma = \pi d n_1 / \lambda \cos \theta` is the
        normalized amplitude of the fundamental grating, :math:`d` is the
        thickness of the grating, :math:`n_1` is the amplitude of the fundamental
        grating, :math:`\lambda` is the free-space wavelength of the incident
        light, and :math:`\theta` is the angle of incidence inside the medium.
        """

        normal_rulings = self.spacing_(rays.position, normal).normalized

        parallel_rulings = normal.cross(normal_rulings).normalized

        direction = rays.direction
        direction = direction - direction @ parallel_rulings

        wavelength = rays.wavelength
        cos_theta = -direction @ normal
        d = self.depth
        i = self.diffraction_order

        gamma = np.pi * d / (wavelength * cos_theta)

        result = scipy.special.jv(i, 2 * gamma)

        return result


@dataclasses.dataclass(eq=False, repr=False)
class SquareRulings(
    AbstractRulings,
):
    r"""
    A ruling profile described by a square wave with a 50% duty cycle.

    Examples
    --------

    Compute the 1st-order groove efficiency of square rulings with a groove
    density of 2500 grooves/mm and a groove depth of 15 nm.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the groove density
        density = 2500 / u.mm

        # Define the groove depth
        depth = 15 * u.nm

        # Define ruling model
        rulings = optika.rulings.SquareRulings(
            spacing=1 / density,
            depth=depth,
            diffraction_order=1,
        )

        # Define the wavelengths at which to sample the groove efficiency
        wavelength = na.geomspace(100, 1000, axis="wavelength", num=1001) * u.AA

        # Define the incidence angles at which to sample the groove efficiency
        angle = na.linspace(0, 30, num=3, axis="angle") * u.deg

        # Define the light rays incident on the grooves
        rays = optika.rays.RayVectorArray(
            wavelength=wavelength,
            direction=na.Cartesian3dVectorArray(
                x=np.sin(angle),
                y=0,
                z=np.cos(angle),
            ),
        )

        # Compute the efficiency of the grooves for the given wavelength
        efficiency = rulings.efficiency(
            rays=rays,
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the groove efficiency as a function of wavelength
        fig, ax = plt.subplots()
        angle_str = angle.value.astype(str).astype(object)
        na.plt.plot(
            wavelength,
            efficiency,
            ax=ax,
            axis="wavelength",
            label=r"$\theta$ = " + angle_str + f"{angle.unit:latex_inline}",
        );
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel(f"efficiency");
        ax.legend();
    """

    spacing: u.Quantity | na.AbstractScalar | AbstractRulingSpacing = MISSING
    """Spacing between adjacent rulings at the given position."""

    depth: u.Quantity | na.AbstractScalar = MISSING
    """Depth of the ruling pattern."""

    diffraction_order: int | na.AbstractScalar = MISSING
    """The diffraction order to simulate."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.spacing),
            optika.shape(self.depth),
            optika.shape(self.diffraction_order),
        )

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> float | na.AbstractScalar:
        r"""
        The fraction of light diffracted into a given order.

        Calculated using the expression given in Table 1 of :cite:t:`Magnusson1978`.

        Parameters
        ----------
        rays
            The light rays incident on the rulings
        normal
            The vector normal to the surface on which the rulings are placed.

        Notes
        -----

        The theoretical efficiency of thin (wavelength much smaller than
        the groove spacing), square rulings is given by Table 1 of
        :cite:t:`Magnusson1978`,

        .. math::

            \eta_i = \begin{cases}
                \cos^2(\pi \gamma / 2) & i = 0 \\
                0 & i = \text{even} \\
                (2 / i \pi)^2 \sin^2 (\pi \gamma / 2) & i = \text{odd}, \\
            \end{cases}

        where :math:`\eta_i` is the groove efficiency for diffraction order
        :math:`i`, :math:`\gamma = \pi d n_1 / \lambda \cos \theta` is the
        normalized amplitude of the fundamental grating, :math:`d` is the
        thickness of the grating, :math:`n_1` is the amplitude of the fundamental
        grating, :math:`\lambda` is the free-space wavelength of the incident
        light, and :math:`\theta` is the angle of incidence inside the medium.
        """

        normal_rulings = self.spacing_(rays.position, normal).normalized

        parallel_rulings = normal.cross(normal_rulings).normalized

        direction = rays.direction
        direction = direction - direction @ parallel_rulings

        wavelength = rays.wavelength
        cos_theta = -direction @ normal
        amplitude = np.pi / 4
        d = self.depth / amplitude
        i = self.diffraction_order

        gamma = np.pi * d / (wavelength * cos_theta)

        result = np.where(
            i % 2 == 0,
            x=0,
            y=np.square(2 * np.sin(np.pi * gamma / 2 * u.rad) / (i * np.pi)),
        )
        result = np.where(
            i == 0,
            x=np.square(np.cos(np.pi * gamma / 2 * u.rad)),
            y=result,
        )

        return result


@dataclasses.dataclass(eq=False, repr=False)
class SawtoothRulings(
    AbstractRulings,
):
    r"""
    A ruling profile described by a sawtooth wave.

    Examples
    --------

    Compute the 1st-order groove efficiency of sawtooth rulings with a groove
    density of 2500 grooves/mm and a groove depth of 15 nm.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the groove density
        density = 2500 / u.mm

        # Define the groove depth
        depth = 15 * u.nm

        # Define ruling model
        rulings = optika.rulings.SawtoothRulings(
            spacing=1 / density,
            depth=depth,
            diffraction_order=-1,
        )

        # Define the wavelengths at which to sample the groove efficiency
        wavelength = na.geomspace(100, 1000, axis="wavelength", num=1001) * u.AA

        # Define the incidence angles at which to sample the groove efficiency
        angle = na.linspace(0, 30, num=3, axis="angle") * u.deg

        # Define the light rays incident on the grooves
        rays = optika.rays.RayVectorArray(
            wavelength=wavelength,
            direction=na.Cartesian3dVectorArray(
                x=np.sin(angle),
                y=0,
                z=np.cos(angle),
            ),
        )

        # Compute the efficiency of the grooves for the given wavelength
        efficiency = rulings.efficiency(
            rays=rays,
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the groove efficiency as a function of wavelength
        fig, ax = plt.subplots()
        angle_str = angle.value.astype(str).astype(object)
        na.plt.plot(
            wavelength,
            efficiency,
            ax=ax,
            axis="wavelength",
            label=r"$\theta$ = " + angle_str + f"{angle.unit:latex_inline}",
        );
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel(f"efficiency");
        ax.legend();
    """

    spacing: u.Quantity | na.AbstractScalar | AbstractRulingSpacing = MISSING
    """Spacing between adjacent rulings at the given position."""

    depth: u.Quantity | na.AbstractScalar = MISSING
    """Depth of the ruling pattern."""

    diffraction_order: int | na.AbstractScalar = MISSING
    """The diffraction order to simulate."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.spacing),
            optika.shape(self.depth),
            optika.shape(self.diffraction_order),
        )

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> float | na.AbstractScalar:
        r"""
        The fraction of light diffracted into a given order.

        Calculated using the expression given in Table 1 of :cite:t:`Magnusson1978`.

        Parameters
        ----------
        rays
            The light rays incident on the rulings
        normal
            The vector normal to the surface on which the rulings are placed.

        Notes
        -----

        The theoretical efficiency of thin (wavelength much smaller than
        the groove spacing), sawtooth rulings is given by Table 1 of
        :cite:t:`Magnusson1978`,

        .. math::

            \eta_i = [\pi (\gamma + i)]^{-2} \sin^2(\pi \gamma)

        where :math:`\eta_i` is the groove efficiency for diffraction order
        :math:`i`, :math:`\gamma = \pi d n_1 / \lambda \cos \theta` is the
        normalized amplitude of the fundamental grating, :math:`d` is the
        thickness of the grating, :math:`n_1` is the amplitude of the fundamental
        grating, :math:`\lambda` is the free-space wavelength of the incident
        light, and :math:`\theta` is the angle of incidence inside the medium.
        """

        normal_rulings = self.spacing_(rays.position, normal).normalized

        parallel_rulings = normal.cross(normal_rulings).normalized

        direction = rays.direction
        direction = direction - direction @ parallel_rulings

        wavelength = rays.wavelength
        cos_theta = -direction @ normal
        amplitude = np.pi / 2
        d = self.depth / amplitude
        i = self.diffraction_order

        gamma = np.pi * d / (wavelength * cos_theta)

        result = np.square(np.sin(np.pi * gamma * u.rad) / (np.pi * (gamma + i)))

        return result


@dataclasses.dataclass(eq=False, repr=False)
class TriangularRulings(
    AbstractRulings,
):
    r"""
    A ruling profile described by a triangle wave.

    Examples
    --------

    Compute the 1st-order groove efficiency of triangular rulings with a groove
    density of 2500 grooves/mm and a groove depth of 15 nm.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the groove density
        density = 2500 / u.mm

        # Define the groove depth
        depth = 15 * u.nm

        # Define ruling model
        rulings = optika.rulings.TriangularRulings(
            spacing=1 / density,
            depth=depth,
            diffraction_order=1,
        )

        # Define the wavelengths at which to sample the groove efficiency
        wavelength = na.geomspace(100, 1000, axis="wavelength", num=1001) * u.AA

        # Define the incidence angles at which to sample the groove efficiency
        angle = na.linspace(0, 30, num=3, axis="angle") * u.deg

        # Define the light rays incident on the grooves
        rays = optika.rays.RayVectorArray(
            wavelength=wavelength,
            direction=na.Cartesian3dVectorArray(
                x=np.sin(angle),
                y=0,
                z=np.cos(angle),
            ),
        )

        # Compute the efficiency of the grooves for the given wavelength
        efficiency = rulings.efficiency(
            rays=rays,
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the groove efficiency as a function of wavelength
        fig, ax = plt.subplots()
        angle_str = angle.value.astype(str).astype(object)
        na.plt.plot(
            wavelength,
            efficiency,
            ax=ax,
            axis="wavelength",
            label=r"$\theta$ = " + angle_str + f"{angle.unit:latex_inline}",
        );
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel(f"efficiency");
        ax.legend();
    """

    spacing: u.Quantity | na.AbstractScalar | AbstractRulingSpacing = MISSING
    """Spacing between adjacent rulings at the given position."""

    depth: u.Quantity | na.AbstractScalar = MISSING
    """Depth of the ruling pattern."""

    diffraction_order: int | na.AbstractScalar = MISSING
    """The diffraction order to simulate."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.spacing),
            optika.shape(self.depth),
            optika.shape(self.diffraction_order),
        )

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> float | na.AbstractScalar:
        r"""
        The fraction of light diffracted into a given order.

        Calculated using the expression given in Table 1 of :cite:t:`Magnusson1978`.

        Parameters
        ----------
        rays
            The light rays incident on the rulings
        normal
            The vector normal to the surface on which the rulings are placed.

        Notes
        -----

        The theoretical efficiency of thin (wavelength much smaller than
        the groove spacing), triangular rulings is given by Table 1 of
        :cite:t:`Magnusson1978`,

        .. math::

            \eta_i = \begin{cases}
                \{\gamma / [(\pi \gamma / 2)^2 - i^2]\}^2 \sin^2 (\pi^2 \gamma / 4) & i = \text{even} \\
                \{\gamma / [(\pi \gamma / 2)^2 - i^2]\}^2 \cos^2 (\pi^2 \gamma / 4) & i = \text{odd}, \\
            \end{cases}

        where :math:`\eta_i` is the groove efficiency for diffraction order
        :math:`i`, :math:`\gamma = \pi d n_1 / \lambda \cos \theta` is the
        normalized amplitude of the fundamental grating, :math:`d` is the
        thickness of the grating, :math:`n_1` is the amplitude of the fundamental
        grating, :math:`\lambda` is the free-space wavelength of the incident
        light, and :math:`\theta` is the angle of incidence inside the medium.
        """

        normal_rulings = self.spacing_(rays.position, normal).normalized

        parallel_rulings = normal.cross(normal_rulings).normalized

        direction = rays.direction
        direction = direction - direction @ parallel_rulings

        wavelength = rays.wavelength
        cos_theta = -direction @ normal
        amplitude = np.square(np.pi) / 8
        d = self.depth / amplitude
        i = self.diffraction_order

        gamma = np.pi * d / (wavelength * cos_theta)

        a = gamma / (np.square(np.pi * gamma / 2) + np.square(i))

        result = np.where(
            i % 2 == 0,
            x=np.square(a * np.sin(np.square(np.pi) * gamma / 4 * u.rad)),
            y=np.square(a * np.cos(np.square(np.pi) * gamma / 4 * u.rad)),
        )

        return result


@dataclasses.dataclass(eq=False, repr=False)
class RectangularRulings(
    AbstractRulings,
):
    r"""
    A ruling profile described by a rectangular wave.

    Examples
    --------

    Compute the 1st-order groove efficiency of rectangular rulings with a groove
    density of 2500 grooves/mm, a groove depth of 15 nm, and a duty cycle of
    30 percent.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the groove density
        density = 2500 / u.mm

        # Define the groove depth
        depth = 15 * u.nm

        # Define ruling model
        rulings = optika.rulings.RectangularRulings(
            spacing=1 / density,
            depth=depth,
            ratio_duty=0.3,
            diffraction_order=1,
        )

        # Define the wavelengths at which to sample the groove efficiency
        wavelength = na.geomspace(100, 1000, axis="wavelength", num=1001) * u.AA

        # Define the incidence angles at which to sample the groove efficiency
        angle = na.linspace(0, 30, num=3, axis="angle") * u.deg

        # Define the light rays incident on the grooves
        rays = optika.rays.RayVectorArray(
            wavelength=wavelength,
            direction=na.Cartesian3dVectorArray(
                x=np.sin(angle),
                y=0,
                z=np.cos(angle),
            ),
        )

        # Compute the efficiency of the grooves for the given wavelength
        efficiency = rulings.efficiency(
            rays=rays,
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the groove efficiency as a function of wavelength
        fig, ax = plt.subplots()
        angle_str = angle.value.astype(str).astype(object)
        na.plt.plot(
            wavelength,
            efficiency,
            ax=ax,
            axis="wavelength",
            label=r"$\theta$ = " + angle_str + f"{angle.unit:latex_inline}",
        );
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel(f"efficiency");
        ax.legend();
    """

    spacing: u.Quantity | na.AbstractScalar | AbstractRulingSpacing = MISSING
    """Spacing between adjacent rulings at the given position."""

    depth: u.Quantity | na.AbstractScalar = MISSING
    """Depth of the ruling pattern."""

    ratio_duty: u.Quantity | na.AbstractScalar = MISSING
    """The duty cycle of the ruling pattern."""

    diffraction_order: int | na.AbstractScalar = MISSING
    """The diffraction order to simulate."""

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.spacing),
            optika.shape(self.depth),
            optika.shape(self.ratio_duty),
            optika.shape(self.diffraction_order),
        )

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> float | na.AbstractScalar:
        r"""
        The fraction of light diffracted into a given order.

        Calculated using the expression given in Table 1 of :cite:t:`Magnusson1978`.

        Parameters
        ----------
        rays
            The light rays incident on the rulings
        normal
            The vector normal to the surface on which the rulings are placed.

        Notes
        -----

        The theoretical efficiency of thin (wavelength much smaller than
        the groove spacing), rectangular rulings is given by Table 1 of
        :cite:t:`Magnusson1978`,

        .. math::

            \eta_i = \begin{cases}
                1 - [(2 a / \pi) - (a / \pi)^2] \sin^2 \left\{ \pi \gamma / [2 (1 - \cos a)]^{1/2} \right\} & i = 0 \\
                [2 / (i \pi)^2](1 - \cos i a) \sin^2 \left\{ \pi \gamma / [2 (1 - \cos a)]^{1/2} \right\} & i \ne 0, \\
            \end{cases}

        where :math:`\eta_i` is the groove efficiency for diffraction order :math:`i`,
        :math:`a` :math:`(0 < a < 2 \pi)` is the duty cycle of the rectangular wave,
        :math:`\gamma = \pi d n_1 / \lambda \cos \theta` is the
        normalized amplitude of the fundamental grating, :math:`d` is the
        thickness of the grating, :math:`n_1` is the amplitude of the fundamental
        grating, :math:`\lambda` is the free-space wavelength of the incident
        light, and :math:`\theta` is the angle of incidence inside the medium.
        """

        normal_rulings = self.spacing_(rays.position, normal).normalized

        parallel_rulings = normal.cross(normal_rulings).normalized

        direction = rays.direction
        direction = direction - direction @ parallel_rulings

        wavelength = rays.wavelength
        cos_theta = -direction @ normal
        a = 2 * np.pi * self.ratio_duty
        amplitude = np.pi / (2 * np.sqrt(2 * (1 - np.cos(a))))
        d = self.depth / amplitude
        i = self.diffraction_order

        gamma = np.pi * d / (wavelength * cos_theta)

        b = np.sin(np.pi * gamma / np.sqrt(2 * (1 - np.cos(a * u.rad))) * u.rad)
        b = np.square(b)

        result = np.where(
            i == 0,
            x=1 - ((2 * a / np.pi) - np.square(a / np.pi)) * b,
            y=(2 / np.square(i * np.pi)) * (1 - np.cos(i * a * u.rad)) * b,
        )

        return result



from __future__ import annotations
from typing import Sequence
import abc
import dataclasses
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import astropy.units as u
import named_arrays as na
import optika
from . import (
    snells_law_scalar,
    matrices,
    AbstractMaterial,
    AbstractMirror,
    AbstractLayer,
    Layer,
    LayerSequence,
)

__all__ = [
    "multilayer_coefficients",
    "multilayer_efficiency",
    "layer_absorbance",
    "AbstractMultilayerMaterial",
    "AbstractMultilayerFilm",
    "MultilayerFilm",
    "AbstractMultilayerMirror",
    "MultilayerMirror",
]


def multilayer_coefficients(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: float | na.AbstractScalar = 1,
    n: float | na.AbstractScalar = 1,
    layers: Sequence[AbstractLayer] | optika.materials.AbstractLayer = None,
    substrate: None | Layer = None,
) -> tuple[
    optika.vectors.PolarizationVectorArray,
    optika.vectors.PolarizationVectorArray,
]:
    r"""
    Calculate the reflection and transmission coefficients of a multilayer
    stack using the method described in :cite:t:`Yeh1988`.

    Parameters
    ----------
    wavelength
        The wavelength of the incident light in vacuum.
    direction
        The component of the incident light's propagation direction in the
        ambient medium antiparallel to the surface normal.
        Default is to assume normal incidence.
    n
        The complex index of refraction of the ambient medium.
    layers
        A sequence of layers representing the multilayer stack.
        If :obj:`None`, then this function computes the reflectivity and
        transmissivity of the ambient medium and the substrate.
    substrate
        A layer representing the substrate supporting the multilayer stack.
        The thickness of this layer is ignored.
        If :obj:`None`, then the substrate is assumed to be a vacuum.

    Notes
    -----

    The reflection and transmission coefficients of the multilayer stack can
    be calculated using the system transfer matrix method described in
    :cite:t:`Yeh1988`.

    The system transfer matrix is calculated using the transfer matrices of each
    layer, where each consists of two parts:
    the refractive matrix and propagation matrix.

    The refractive matrix is given by :cite:t:`Yeh1988` Equation 5.1-12,

    .. math::
        :label: refractive-matrix

        W_{kij} = \frac{1}{t_{kij}} \begin{pmatrix}
                                    1 & r_{kij} \\
                                    r_{kij} & 1 \\
                                  \end{pmatrix},

    where :math:`k=(s, p)` is the polarization state, :math:`i=j-1` is the index
    of the previous material, :math:`j` is the index of the current material,

    .. math::
        :label: fresnel-reflection

        r_{kij} = \frac{q_{ki} - q_{kj}}{q_{ki} + q_{kj}}

    is the Fresnel reflection coefficient between materials :math:`i` and :math:`j`,

    .. math::
        :label: fresnel-transmission

        t_{kij} = \frac{2 q_{ki}}{q_{ki} + q_{kj}}

    is the Fresnel transmission coefficient between materials :math:`i` and :math:`j`,

    .. math::

        q_{si} = n_i \cos \theta_i

    and

    .. math::

        q_{pi} = \frac{\cos \theta_i}{n_i}

    are the :math:`z` components of the wave's momentum for :math:`s` and
    :math:`p` polarization,
    :math:`n_i` is the index of refraction inside material :math:`i`,
    and :math:`\theta_i` is the angle between the wave's propagation direction
    and the vector normal to the interface inside material :math:`i`.

    The propagation matrix takes the form

    .. math::
        :label: propagation-matrix

        U_{kj} = \begin{pmatrix}
                    e^{-i \beta_j} & 0 \\
                    0 & e^{i \beta_j} \\
                \end{pmatrix}

    where

    .. math::

        \beta_j = \frac{2 \pi}{\lambda} n_j h_j \cos \theta_j,

    is the phase change from propagating through material :math:`j`,
    :math:`\lambda` is the vacuum wavelength of the incident light,
    and :math:`h_j` is the thickness of material :math:`j`.

    To compute the system transfer matrix, we find the matrix product of the
    :math:`S=N+1` refractive matrices from each interface and the :math:`N`
    propagation matrices from each layer

    .. math::
        :label: sytem-transfer-matrix

        M_k = \left( \prod_{j=1}^N W_{kij} U_{kj} \right) W_{kNS}

    where :math:`i=j-1`.

    Once the system transfer matrix has been calculated, we can use
    :cite:t:`Yeh1988` Equation 5.2-3 to compute the system reflection coefficient

    .. math::
        :label: system-fresnel-reflection

        r_k = \frac{M_{k21}}{M_{k11}},

    and Equation 5.2-4 to compute the system transmission coefficient

    .. math::
        :label: system-fresnel-transmission

        t_k = \frac{1}{M_{k11}}.

    If we define the vectors

    .. math::

        \vec{r} = \begin{pmatrix}
            r_s \\
            r_p
        \end{pmatrix}

    and

    .. math::

        \vec{t} = \begin{pmatrix}
            t_s \\
            t_p
        \end{pmatrix},

    then the :class:`tuple` :math:`(\vec{r}, \vec{t})` is the quantity returned
    by this function.
    """
    direction_ambient = direction
    n_ambient = n

    if substrate is not None:
        substrate = dataclasses.replace(substrate, thickness=0 * u.nm)
    else:
        substrate = optika.materials.Layer(thickness=0 * u.nm)

    polarized_s = na.ScalarArray(
        ndarray=np.array([True, False]),
        axes="_polarization",
    )

    if layers is None:
        layers = []
    if not isinstance(layers, optika.materials.AbstractLayer):
        layers = optika.materials.LayerSequence(layers)

    n, direction, m_layers, where = layers.transfer(
        wavelength=wavelength,
        direction=direction_ambient,
        polarized_s=polarized_s,
        n=n_ambient,
    )
    n, direction, m_substrate, where = substrate.transfer(
        wavelength=wavelength,
        direction=direction,
        polarized_s=polarized_s,
        n=n,
        where=where,
    )
    m = m_layers @ m_substrate

    r = m.y.x / m.x.x
    t = 1 / m.x.x
    t[~where] = 0

    index_s = dict(_polarization=0)
    index_p = dict(_polarization=1)

    r = optika.vectors.PolarizationVectorArray(
        s=r[index_s],
        p=r[index_p],
    )

    t = optika.vectors.PolarizationVectorArray(
        s=t[index_s],
        p=t[index_p],
    )

    return r, t


def multilayer_efficiency(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: float | na.AbstractScalar = 1,
    n: float | na.AbstractScalar = 1,
    layers: None | Sequence[AbstractLayer] | optika.materials.AbstractLayer = None,
    substrate: None | Layer = None,
) -> tuple[
    optika.vectors.PolarizationVectorArray,
    optika.vectors.PolarizationVectorArray,
]:
    r"""
    Calculate the reflectivity and transmissivity of a multilayer
    film or coating using the method in :cite:t:`Windt1998`.

    Parameters
    ----------
    wavelength
        The wavelength of the incident light in vacuum.
    direction
        The component of the incident light's propagation direction in the
        ambient medium antiparallel to the surface normal.
        Default is to assume normal incidence.
    n
        The complex index of refraction of the ambient medium.
    layers
        A sequence of layers representing the multilayer stack.
        If :obj:`None`, then this function computes the reflectivity and
        transmissivity of the ambient medium and the substrate.
    substrate
        A layer representing the substrate supporting the multilayer stack.
        The thickness of this layer is ignored.
        If :obj:`None`, then the substrate is assumed to be a vacuum.

    Examples
    --------

    Reproduce Example 2.3.1 in the
    `IMD User's Manual <http://www.rxollc.com/idl/IMD.pdf>`_,
    the transmittance of a :math:`\text{Zr}` filter.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the wavelength of the incident light
        wavelength = na.linspace(100, 150, axis="wavelength", num=501) * u.AA

        # Define the Zr layer
        layers = optika.materials.Layer(
            chemical="Zr",
            thickness=1500 * u.AA,
        )

        # Compute the reflectivity and the transmissivity of this multilayer
        reflectivity, transmissivity = optika.materials.multilayer_efficiency(
            wavelength=wavelength,
            layers=layers,
        )

        # Plot the transmissivity as a function of wavelength.
        fig, ax = plt.subplots()
        na.plt.plot(
            wavelength,
            transmissivity.average,
            ax=ax,
            axis="wavelength",
            label="Zr",
        );
        ax.legend();
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("transmissivity");

    Reproduce Example 2.3.2 in the
    `IMD User's Manual <http://www.rxollc.com/idl/IMD.pdf>`_,
    the reflectivity of a :math:`\text{Si/Mo}` multilayer stack with
    interfacial roughness.

    .. jupyter-execute::

        # Define the wavelength of the incident light
        wavelength = na.linspace(100, 150, axis="wavelength", num=501) * u.AA

        # Period length of the multilayer sequence
        d = 66.5 * u.AA

        # Define the thickness to period ratios for each layer
        thickness_ratio = 0.6

        # Define the interface profile between successive layers
        interface = optika.materials.profiles.ErfInterfaceProfile(7 * u.AA)

        # Define the multilayer sequence
        layers = optika.materials.PeriodicLayerSequence(
            [
                optika.materials.Layer(
                    chemical="Si",
                    thickness=thickness_ratio * d,
                    interface=interface,
                ),
                optika.materials.Layer(
                    chemical="Mo",
                    thickness=(1 - thickness_ratio) * d,
                    interface=interface,
                ),
            ],
            num_periods=60,
        )

        # Compute the reflectivity and transmissivity of this multilayer stack
        reflectivity, transmissivity = optika.materials.multilayer_efficiency(
            wavelength=wavelength,
            layers=layers,
        )

        # Plot the reflectivity as a function of wavelength
        fig, ax = plt.subplots()
        na.plt.plot(
            wavelength,
            reflectivity.average,
            ax=ax,
            axis="wavelength",
            label=rf"Si/Mo $\times$ {layers.num_periods}",
        );
        ax.legend();
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("reflectivity");

    Reproduce Figure 9a in :cite:t:`Windt1998`, the reflectivity of a
    :math:`\text{Y/Al}` multilayer stack as a function of wavelength and of
    the ratio of the :math:`\text{Y}` thickness to the :math:`\text{Y + Al}`
    thickness, :math:`\Gamma`.

    .. jupyter-execute::

        # Period length of the multilayer sequence
        d = 98 * u.AA

        # wavelength of the incident light
        wavelength = na.linspace(170, 210, num=101, axis="wavelength") * u.AA

        # an array of thickness-to-period ratios for each layer
        thickness_ratio = na.linspace(0.2, 0.6, axis="thickness_ratio", num=5)

        # Define the multilayer sequence
        layers = optika.materials.PeriodicLayerSequence(
            [
                optika.materials.Layer(
                    chemical="Y",
                    thickness=thickness_ratio * d,
                ),
                optika.materials.Layer(
                    chemical="Al",
                    thickness=(1 - thickness_ratio) * d,
                )
            ],
            num_periods=40,
        )

        # Define the substrate layer
        substrate = optika.materials.Layer(
            chemical="Si",
        )

        # Compute the reflectivity and transmissivity of this multilayer stack
        reflectivity, transmissivity = optika.materials.multilayer_efficiency(
            wavelength,
            layers=layers,
            substrate=substrate,
        )

        # Plot the reflectivity as a function of wavelength
        fig, ax = plt.subplots()
        na.plt.plot(
            wavelength,
            reflectivity.average,
            ax=ax,
            axis="wavelength",
            label=r"$\Gamma=" + thickness_ratio.astype(str).astype(object) + "$",
        );
        ax.legend();
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("reflectivity");

    Notes
    -----

    The reflectivity of a multilayer stack can be calculated
    using Equation 5.2-5 in :cite:t:`Yeh1988`,

    .. math::
        :label: reflectivity

        R_k = |r_k|^2,

    and the transmissivity can be calculated using Equation 5.2-6,

    .. math::
        :label: transmissivity

        T_k = \text{Re} \left( \frac{q_{kS}}{q_{k0}} \right) |t_k|^2,

    where :math:`r_k` and :math:`t_k` are the system reflection and transmission
    coefficients calculated by :func:`multilayer_coefficients`,
    :math:`k = (s, p)` is the polarization state,

    .. math::

        q_{si} = n_i \cos \theta_i

    and

    .. math::

        q_{pi} = \frac{\cos \theta_i}{n_i}

    are the :math:`z` components of the wave's momentum for an arbitrary layer :math:`i`,
    :math:`n_i` is the index of refraction inside material :math:`i`,
    and :math:`\theta_i` is the angle between the wave's propagation direction
    and the vector normal to the interface inside material :math:`i`.

    If we define the vectors

    .. math::

        \vec{R} = \begin{pmatrix}
            R_s \\
            R_p
        \end{pmatrix}

    and

    .. math::

        \vec{T} = \begin{pmatrix}
            T_s \\
            T_p
        \end{pmatrix},

    then the :class:`tuple` :math:`(\vec{R}, \vec{T})` is the quantity returned
    by this function.
    """
    direction_ambient = direction
    n_ambient = n

    if substrate is not None:
        substrate = dataclasses.replace(substrate, thickness=0 * u.nm)
    else:
        substrate = optika.materials.Layer(thickness=0 * u.nm)

    r, t = multilayer_coefficients(
        wavelength=wavelength,
        direction=direction_ambient,
        n=n_ambient,
        layers=layers,
        substrate=substrate,
    )

    n_substrate = substrate.n(wavelength)
    direction_substrate = snells_law_scalar(
        cos_incidence=direction_ambient,
        index_refraction=n_ambient,
        index_refraction_new=n_substrate,
    )

    impedance_ambient = optika.vectors.PolarizationVectorArray(
        s=n_ambient,
        p=1 / n_ambient,
    )
    impedance_substrate = optika.vectors.PolarizationVectorArray(
        s=n_substrate,
        p=1 / n_substrate,
    )

    direction_ambient = optika.vectors.PolarizationVectorArray(
        s=direction_ambient,
        p=np.conj(direction_ambient),
    )
    direction_substrate = optika.vectors.PolarizationVectorArray(
        s=direction_substrate,
        p=np.conj(direction_substrate),
    )

    q_ambient = direction_ambient * impedance_ambient
    q_substrate = direction_substrate * impedance_substrate

    reflectivity = np.square(np.abs(r))
    transmissivity = np.square(np.abs(t)) * np.real(q_substrate / q_ambient)

    return reflectivity, transmissivity


def layer_absorbance(
    index: int,
    wavelength: u.Quantity | na.AbstractScalar,
    direction: float | na.AbstractScalar = 1,
    n: float | na.AbstractScalar = 1,
    layers: None | Sequence[AbstractLayer] | optika.materials.AbstractLayer = None,
    substrate: None | Layer = None,
) -> optika.vectors.PolarizationVectorArray:
    """
    Compute the fraction of energy absorbed for a particular layer in the
    multilayer stack.

    Parameters
    ----------
    index
        The index of a :class:`AbstractLayer` in `layers` to find the absorbance of.
    wavelength
        The wavelength of the incident light in vacuum.
    direction
        The component of the incident light's propagation direction in the
        ambient medium antiparallel to the surface normal.
        Default is to assume normal incidence.
    n
        The complex index of refraction of the ambient medium.
    layers
        A sequence of layers representing the multilayer stack.
        If :obj:`None`, then this function computes the reflectivity and
        transmissivity of the ambient medium and the substrate.
    substrate
        A layer representing the substrate supporting the multilayer stack.
        The thickness of this layer is ignored.
        If :obj:`None`, then the substrate is assumed to be a vacuum.

    Examples
    --------

    Compute the amount of energy absorbed in a layer of silicon coated by a
    thin layer of silicon dioxide and compare it to the energy absorbed by
    both layers

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define a grid of wavelengths at which to evaluate the absorbance
        wavelength = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Define a thin layer of silicon dioxide and a thick layer of silicon
        layers = [
            optika.materials.Layer(
                chemical="SiO2",
                thickness=5 * u.nm,
            ),
            optika.materials.Layer(
                chemical="Si",
                thickness=16 * u.um,
            )
        ]

        # Compute the fraction of energy absorbed by the silicon
        absorbance = optika.materials.layer_absorbance(
            index=1,
            wavelength=wavelength,
            layers=layers,
        )

        # Compute the fraction of energy absorbed by both layers
        reflected, transmitted = optika.materials.multilayer_efficiency(
            wavelength=wavelength,
            layers=layers,
        )
        absorbance_total = 1 - reflected - transmitted

        # Plot the results
        fig, ax = plt.subplots()
        na.plt.plot(
            wavelength,
            absorbance.average,
            ax=ax,
            label="absorbance",
        )
        na.plt.plot(
            wavelength,
            absorbance_total.average,
            ax=ax,
            label="total absorbance",
        )
        ax.set_xscale("log")
        ax.legend()
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("incident energy fraction");
    """
    if not isinstance(layers, AbstractLayer):
        layers = LayerSequence(layers)
    layers = layers.layer_sequence

    if substrate is not None:
        substrate = dataclasses.replace(substrate, thickness=0 * u.nm)
    else:
        substrate = optika.materials.Layer(thickness=0 * u.nm)

    r, t = multilayer_coefficients(
        wavelength=wavelength,
        direction=direction,
        n=n,
        layers=layers,
        substrate=substrate,
    )
    r = na.stack([r.s, r.p], axis="_polarization")

    layer = layers[index]

    polarized_s = na.ScalarArray(
        ndarray=np.array([True, False]),
        axes="_polarization",
    )

    n_i, direction_i, m_i, where_i = layers[:index].transfer(
        wavelength=wavelength,
        direction=direction,
        polarized_s=polarized_s,
        n=n,
    )

    n_left = layer.n(wavelength)
    direction_left = snells_law_scalar(direction, n, n_left)
    m_left = m_i @ matrices.refraction(
        wavelength=wavelength,
        direction_left=direction_i,
        direction_right=direction_left,
        polarized_s=polarized_s,
        n_left=n_i,
        n_right=n_left,
        interface=layer.interface,
    )

    n_right, direction_right, m_right, where_right = layers[: index + 1].transfer(
        wavelength=wavelength,
        direction=direction,
        polarized_s=polarized_s,
        n=n,
    )

    amplitude_ambient = na.Cartesian2dVectorArray(1, r)

    a_right = m_right.inverse @ amplitude_ambient
    a_left = m_left.inverse @ amplitude_ambient

    a_right[~where_right] = 0

    a_right, b_right = a_right.x, a_right.y
    a_left, b_left = a_left.x, a_left.y

    angle_right = np.arccos(direction_right)
    angle_left = np.arccos(direction_left)

    rotation_right = na.Cartesian3dYRotationMatrixArray(angle_right)
    rotation_left = na.Cartesian3dYRotationMatrixArray(angle_left)

    antirotation_right = na.Cartesian3dYRotationMatrixArray(-angle_right)
    antirotation_left = na.Cartesian3dYRotationMatrixArray(-angle_left)

    a_right = a_right * np.where(
        polarized_s,
        na.Cartesian3dVectorArray(0, 1, 0),
        rotation_right @ na.Cartesian3dVectorArray(1, 0, 0),
    )
    a_left = a_left * np.where(
        polarized_s,
        na.Cartesian3dVectorArray(0, 1, 0),
        rotation_left @ na.Cartesian3dVectorArray(1, 0, 0),
    )

    b_right = b_right * np.where(
        polarized_s,
        na.Cartesian3dVectorArray(0, 1, 0),
        antirotation_right @ na.Cartesian3dVectorArray(1, 0, 0),
    )
    b_left = b_left * np.where(
        polarized_s,
        na.Cartesian3dVectorArray(0, 1, 0),
        antirotation_left @ na.Cartesian3dVectorArray(1, 0, 0),
    )

    impedance_ambient = np.where(
        polarized_s,
        n,
        1 / n,
    )
    impedance_right = np.where(
        polarized_s,
        n_right,
        1 / n_right,
    )
    impedance_left = np.where(
        polarized_s,
        n_left,
        1 / n_left,
    )

    ka_right = rotation_right @ na.Cartesian3dVectorArray(0, 0, 1)
    ka_left = rotation_left @ na.Cartesian3dVectorArray(0, 0, 1)
    kb_right = antirotation_right @ na.Cartesian3dVectorArray(0, 0, -1)
    kb_left = antirotation_left @ na.Cartesian3dVectorArray(0, 0, -1)

    ka_right = impedance_right * ka_right
    ka_left = impedance_left * ka_left
    kb_right = impedance_right * kb_right
    kb_left = impedance_left * kb_left

    c_right = ka_right.cross(a_right)
    c_left = ka_left.cross(a_left)

    d_right = kb_right.cross(b_right)
    d_left = kb_left.cross(b_left)

    flux_right = np.real((a_right + b_right).cross(np.conj(c_right + d_right)))
    flux_left = np.real((a_left + b_left).cross(np.conj(c_left + d_left)))

    flux_absorbed = (flux_left.z - flux_right.z) / (impedance_ambient * direction)

    index_s = dict(_polarization=0)
    index_p = dict(_polarization=1)

    flux_absorbed = optika.vectors.PolarizationVectorArray(
        s=flux_absorbed[index_s],
        p=flux_absorbed[index_p],
    )

    return flux_absorbed


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMultilayerMaterial(
    AbstractMaterial,
):
    @property
    @abc.abstractmethod
    def layers(self) -> AbstractLayer | Sequence[AbstractLayer]:
        """
        A sequence of layers representing the multilayer stack.
        """

    @property
    def _layers(self) -> AbstractLayer:
        result = self.layers
        if not isinstance(result, AbstractLayer):
            result = LayerSequence(result)
        return result

    @property
    def transformation(self) -> None:
        return None

    def index_refraction(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return rays.index_refraction

    def attenuation(
        self,
        rays: optika.rays.RayVectorArray,
    ) -> na.ScalarLike:
        return rays.attenuation

    @abc.abstractmethod
    def plot_layers(
        self,
        width: u.Quantity = 100 * u.nm,
        ax: matplotlib.axes.Axes = None,
        **kwargs,
    ) -> list[matplotlib.patches.Polygon]:
        """
        Plot the multilayer stack using :meth:`optika.materials.AbstractLayer.plot`.

        Parameters
        ----------
        width
            The width of the plotted multilayer stack in physical units.
        ax
            The matplotlib axes on which to plot the multilayer stack.
        kwargs
            Additional keyword arguments to pass along to
            :meth:`optika.materials.AbstractLayer.plot`.
        """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMultilayerFilm(
    AbstractMultilayerMaterial,
):
    @property
    def is_mirror(self) -> bool:
        return False

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        """
        Compute the efficiency of this multilayer film using
        :func:`optika.materials.multilayer_efficiency`.

        Parameters
        ----------
        rays
            the input rays with which to compute the efficiency of the multilayer film
        normal
            the vector normal to the interface between successive layers
        """
        wavelength = rays.wavelength
        k = rays.attenuation * wavelength / (4 * np.pi)
        n = rays.index_refraction + k * 1j

        reflectivity, transmissivity = multilayer_efficiency(
            wavelength=wavelength,
            direction=-rays.direction @ normal,
            n=n,
            layers=self.layers,
            substrate=None,
        )
        return transmissivity.average

    def plot_layers(
        self,
        width: u.Quantity = 100 * u.nm,
        ax: matplotlib.axes.Axes = None,
        **kwargs,
    ) -> list[matplotlib.patches.Polygon]:
        return self._layers.plot(
            width=width,
            ax=ax,
            **kwargs,
        )


@dataclasses.dataclass(eq=False, repr=False)
class MultilayerFilm(
    AbstractMultilayerFilm,
):
    layers: None | AbstractLayer | Sequence[AbstractLayer] = None
    """A sequence of layers representing the multilayer stack."""

    @property
    def shape(self) -> dict[str, int]:
        try:
            shape_layer = na.broadcast_shapes(
                *[optika.shape(layer) for layer in self.layers],
            )
        except TypeError:
            shape_layer = optika.shape(self.layers)
        return shape_layer


@dataclasses.dataclass(eq=False, repr=False)
class AbstractMultilayerMirror(
    AbstractMultilayerMaterial,
    AbstractMirror,
):

    def efficiency(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        """
        Compute the efficiency of this multilayer film using
        :func:`optika.materials.multilayer_efficiency`.

        Parameters
        ----------
        rays
            the input rays with which to compute the efficiency of the multilayer film
        normal
            the vector normal to the interface between successive layers
        """
        wavelength = rays.wavelength
        k = rays.attenuation * wavelength / (4 * np.pi)
        n = rays.index_refraction + k * 1j

        reflectivity, transmissivity = multilayer_efficiency(
            wavelength=wavelength,
            direction=-rays.direction @ normal,
            n=n,
            layers=self.layers,
            substrate=self.substrate,
        )
        return reflectivity.average

    def plot_layers(
        self,
        width: u.Quantity = 100 * u.nm,
        ax: matplotlib.axes.Axes = None,
        thickness_substrate: u.Quantity = 10 * u.nm,
        **kwargs,
    ) -> list[matplotlib.patches.Polygon]:

        if ax is None:
            ax = plt.gca()

        result = self._layers.plot(
            width=width,
            ax=ax,
            **kwargs,
        )

        substrate = self.substrate

        amplitude = 0.1 * thickness_substrate
        x_substrate = np.linspace(0, width, num=1001)
        y_substrate = amplitude * np.sin(2 * np.pi * x_substrate / width * u.rad)
        y_substrate = y_substrate - thickness_substrate

        kwargs_substrate = substrate.kwargs_plot
        if kwargs_substrate is None:
            kwargs_substrate = dict()

        ax.fill_between(
            x=x_substrate,
            y1=0,
            y2=y_substrate,
            **(kwargs_substrate | kwargs),
        )

        formula = substrate._chemical.formula_latex
        thickness = substrate.thickness

        ax.text(
            x=width / 2,
            y=-thickness_substrate / 2,
            s=rf"{formula} (${thickness.value:0.0f}\,${thickness.unit:latex_inline})",
            ha="center",
            va="center",
        )

        return result


@dataclasses.dataclass(eq=False, repr=False)
class MultilayerMirror(
    AbstractMultilayerMirror,
):
    r"""
    A model of a mirror coating consisting of alternating layers of different
    materials.

    Examples
    --------

    Reproduce Example 2.3.2 in the
    `IMD User's Manual <http://www.rxollc.com/idl/IMD.pdf>`_,
    the reflectivity of a :math:`\text{Si/Mo}` multilayer stack with
    interfacial roughness.

    .. jupyter-execute::

        import numpy as np
        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Period length of the multilayer sequence
        d = 66.5 * u.AA

        # Number of periods
        N = 60

        # Define the thickness to period ratios for each layer
        thickness_ratio = 0.6

        # Define the interface profile between successive layers
        interface = optika.materials.profiles.ErfInterfaceProfile(
            width=7 * u.AA,
        )

        # Define the periodic sequence of layers
        layers = optika.materials.PeriodicLayerSequence(
            [
                optika.materials.Layer(
                    chemical="Si",
                    thickness=thickness_ratio * d,
                    interface=interface,
                    kwargs_plot=dict(
                        color="tab:blue",
                        alpha=0.5,
                    ),
                ),
                optika.materials.Layer(
                    chemical="Mo",
                    thickness=(1- thickness_ratio) * d,
                    interface=interface,
                    kwargs_plot=dict(
                        color="tab:orange",
                        alpha=0.5,
                    ),
                ),
            ],
            num_periods=60,
        )

        # Define the substrate layer
        substrate = optika.materials.Layer(
            chemical="SiO2",
            thickness=10 * u.mm,
            interface=interface,
            kwargs_plot=dict(
                color="gray",
                alpha=0.5,
            ),
        )

        # Define a representation of multilayer coating
        multilayer = optika.materials.MultilayerMirror(
            layers=layers,
            substrate=substrate,
        )

        # Define the wavelengths of the incident light
        wavelength = na.linspace(100, 150, axis="wavelength", num=501) * u.AA

        # Define the rays incident on the multilayer coating
        rays = optika.rays.RayVectorArray(
            wavelength=wavelength,
            direction=na.Cartesian3dVectorArray(0, 0, 1),
        )

        # Compute the reflectivity of this multilayer coating
        reflectivity = multilayer.efficiency(
            rays=rays,
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the reflectivity as a function of wavelength
        fig, ax = plt.subplots()
        na.plt.plot(
            wavelength,
            reflectivity,
            ax=ax,
            axis="wavelength",
            label=rf"Si/Mo $\times$ {N}",
        );
        ax.legend();
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("reflectivity");

    Plot a visual representation of the multilayer coating

    .. jupyter-execute::

        fig, ax = plt.subplots(constrained_layout=True)
        ax.set_axis_off()
        with astropy.visualization.quantity_support():
            multilayer.plot_layers()
    """

    layers: None | AbstractLayer | Sequence[AbstractLayer] = None
    """A sequence of layers representing the multilayer stack."""

    substrate: None | Layer = None
    """A layer representing the substrate that the layers are deposited onto."""

    @property
    def shape(self) -> dict[str, int]:
        try:
            shape_layer = na.broadcast_shapes(
                *[optika.shape(layer) for layer in self.layers],
            )
        except TypeError:
            shape_layer = optika.shape(self.layers)
        return na.broadcast_shapes(
            shape_layer,
            optika.shape(self.substrate),
        )



from typing import TypeVar, Generic
import abc
import dataclasses
import numpy as np
import numpy.typing as npt
import matplotlib.axes
import named_arrays as na
import optika

__all__ = [
    "AbstractSurface",
    "Surface",
]

SagT = TypeVar(
    "SagT",
    bound=None | optika.sags.AbstractSag,
)
MaterialT = TypeVar(
    "MaterialT",
    bound=None | optika.materials.AbstractMaterial,
)
ApertureT = TypeVar(
    "ApertureT",
    bound=None | optika.apertures.AbstractAperture,
)
ApertureMechanicalT = TypeVar(
    "ApertureMechanicalT",
    bound=None | optika.apertures.AbstractAperture,
)
RulingsT = TypeVar(
    "RulingsT",
    bound=None | optika.rulings.AbstractRulings,
)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSurface(
    optika.mixins.Plottable,
    optika.mixins.Printable,
    optika.mixins.Transformable,
    optika.mixins.Shaped,
    optika.propagators.AbstractLightPropagator,
    Generic[SagT, MaterialT, ApertureT, ApertureMechanicalT, RulingsT],
):
    """
    Interface for a single optical surface.
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        human-readable name of the surface
        """

    @property
    @abc.abstractmethod
    def sag(self) -> SagT:
        """
        the sag profile of the surface
        """

    @property
    @abc.abstractmethod
    def material(self) -> MaterialT:
        """
        optical material type of the surface
        """

    @property
    @abc.abstractmethod
    def aperture(self) -> ApertureT:
        """
        region of the surface which allows light to propagate
        """

    @property
    @abc.abstractmethod
    def aperture_mechanical(self) -> ApertureMechanicalT:
        """
        shape of the physical substrate containing the optical surface
        """

    @property
    @abc.abstractmethod
    def rulings(self) -> RulingsT:
        """
        the ruling density of the surface
        """

    @property
    @abc.abstractmethod
    def is_field_stop(self) -> bool:
        """
        flag controlling whether this surface should act as the field stop
        for the system
        """

    @property
    @abc.abstractmethod
    def is_pupil_stop(self) -> bool:
        """
        flag controlling whether this surface should act as the pupil stop
        for the system
        """

    @property
    def is_stop(self) -> bool:
        """
        If this surface is pupil stop or the field stop, return :obj:`True`.
        """
        return self.is_field_stop or self.is_pupil_stop

    def propagate_rays(
        self,
        rays: optika.rays.RayVectorArray,
        # material: None | optika.materials.AbstractMaterial = None,
    ) -> optika.rays.RayVectorArray:
        r"""
        Refract, reflect, and/or diffract the given rays off of this surface

        Parameters
        ----------
        rays
            a set of input rays that will interact with this surface
        """
        sag = self.sag
        material = self.material
        aperture = self.aperture
        rulings = self.rulings
        transformation = self.transformation

        if transformation is not None:
            rays = transformation.inverse(rays)

        rays_1 = sag.intercept(rays)

        wavelength_1 = rays_1.wavelength
        position_1 = rays_1.position
        a = rays_1.direction
        attenuation_1 = rays_1.attenuation
        displacement = position_1 - rays.position
        intensity_1 = rays_1.intensity * np.exp(-attenuation_1 * displacement.length)
        n1 = rays_1.index_refraction

        normal = sag.normal(position_1)

        position_2 = position_1
        n2 = material.index_refraction(rays_1)
        r = n1 / n2

        wavelength_2 = wavelength_1 / r

        if rulings is not None:
            diffraction_order = rulings.diffraction_order
            vector_rulings = rulings.spacing_(position_1, normal)
            spacing_rulings = vector_rulings.length
            normal_rulings = vector_rulings / spacing_rulings
        else:
            diffraction_order = 0
            spacing_rulings = None
            normal_rulings = None

        b = optika.materials.snells_law(
            wavelength=wavelength_1,
            direction=a,
            index_refraction=n1,
            index_refraction_new=n2,
            is_mirror=material.is_mirror,
            normal=normal,
            diffraction_order=diffraction_order,
            spacing_rulings=spacing_rulings,
            normal_rulings=normal_rulings,
        )

        efficiency = material.efficiency(rays_1, normal)
        if rulings is not None:
            efficiency = efficiency * rulings.efficiency(rays_1, normal)

        intensity_2 = intensity_1 * efficiency
        attenuation_2 = material.attenuation(rays_1)

        rays_2 = dataclasses.replace(
            rays_1,
            wavelength=wavelength_2,
            position=position_2,
            direction=b,
            intensity=intensity_2,
            attenuation=attenuation_2,
            index_refraction=n2,
        )

        if aperture is not None:
            rays_2 = aperture.clip_rays(rays_2)

        if transformation is not None:
            rays_2 = transformation(rays_2)

        return rays_2

    def plot(
        self,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray] = None,
        transformation: None | na.transformations.AbstractTransformation = None,
        components: None | tuple[str, ...] = None,
        **kwargs,
    ) -> dict[str, na.AbstractScalar]:
        sag = self.sag
        aperture = self.aperture
        aperture_mechanical = self.aperture_mechanical
        transformation_self = self.transformation
        kwargs_plot = self.kwargs_plot

        if transformation is not None:
            if transformation_self is not None:
                transformation = transformation @ transformation_self
        else:
            if transformation_self is not None:
                transformation = transformation_self

        if kwargs_plot is not None:
            kwargs = kwargs | kwargs_plot

        result = dict()

        if aperture is not None:
            result["aperture"] = aperture.plot(
                ax=ax,
                transformation=transformation,
                components=components,
                sag=sag,
                **kwargs,
            )

        if aperture_mechanical is not None:
            result["aperture_mechanical"] = aperture_mechanical.plot(
                ax=ax,
                transformation=transformation,
                components=components,
                sag=sag,
                **kwargs,
            )

        return result


@dataclasses.dataclass(eq=False, repr=False)
class Surface(
    AbstractSurface[SagT, MaterialT, ApertureT, ApertureMechanicalT, RulingsT],
):
    """
    Representation of a single optical interface.

    Composition of a sag profile, material type, aperture, and ruling specification (all optional).

    Examples
    --------

    Define a spherical mirror, with a rectangular aperture, :math:`z=50 \\; \\text{mm}` from the origin.

    Reflect a grid of collimated rays off of this mirror and measure their position
    at the origin.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # define a spherical reflective surface 50 mm from the origin
        mirror = optika.surfaces.Surface(
            sag=optika.sags.SphericalSag(radius=-100 * u.mm),
            material=optika.materials.Mirror(),
            aperture=optika.apertures.RectangularAperture(30 * u.mm),
            transformation=na.transformations.Cartesian3dTranslation(z=50 * u.mm),
        )

        # define a detector surface at the origin to capture the reflected rays
        detector=optika.surfaces.Surface()

        # define a grid of collimated input rays
        rays_input = optika.rays.RayVectorArray(
            position=na.Cartesian3dVectorArray(
                x=na.linspace(-25, 25, axis="pupil_x", num=5) * u.mm,
                y=na.linspace(-25, 25, axis="pupil_y", num=5) * u.mm,
                z=0 * u.mm,
            ),
            direction=na.Cartesian3dVectorArray(0, 0, 1),
        )

        # propagate the rays through the mirror and detector surfaces
        rays_mirror = mirror.propagate_rays(rays_input)
        rays_detector = detector.propagate_rays(rays_mirror)

        # stack the 3 sets of rays into a single object
        # for easier plotting
        rays = [
            rays_input,
            rays_mirror,
            rays_detector,
        ]
        rays = na.stack(rays, axis="surface")

        # plot the rays and surface
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            ax.set_aspect("equal")
            components_plot = ("z", "y")
            na.plt.plot(rays.position, axis="surface", components=components_plot, color="tab:blue");
            mirror.plot(ax=ax, components=components_plot, color="black");
    """

    name: None | str = None
    sag: SagT = None
    material: MaterialT = None
    aperture: ApertureT = None
    aperture_mechanical: ApertureMechanicalT = None
    rulings: RulingsT = None
    is_field_stop: bool = False
    is_pupil_stop: bool = False
    transformation: None | na.transformations.AbstractTransformation = None
    kwargs_plot: None | dict = None

    def __post_init__(self):
        if self.sag is None:
            self.sag = optika.sags.NoSag()
        if self.material is None:
            self.material = optika.materials.Vacuum()

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.name),
            optika.shape(self.sag),
            optika.shape(self.material),
            optika.shape(self.aperture),
            optika.shape(self.rulings),
            optika.shape(self.transformation),
        )



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from optika._tests import test_mixins


class AbstractTestAbstractInterfaceProfile(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestShaped,
):
    def test_width(self, a: optika.materials.profiles.AbstractInterfaceProfile):
        result = a.width
        assert na.unit_normalized(result).is_equivalent(u.mm)

    @pytest.mark.parametrize(
        argnames="z",
        argvalues=[
            0 * u.nm,
            na.linspace(-1, 1, axis="z", num=11) * u.nm,
            na.NormalUncertainScalarArray(0, width=1, num_distribution=11) * u.nm,
        ],
    )
    def test__call__(
        self,
        a: optika.materials.profiles.AbstractInterfaceProfile,
        z: u.Quantity | na.AbstractScalar,
    ):
        result = a(z)
        assert na.unit_normalized(result).is_equivalent(u.dimensionless_unscaled)
        assert np.all(result >= 0)
        assert np.all(result <= 1)

    @pytest.mark.parametrize(
        argnames="wavelength",
        argvalues=[
            304 * u.AA,
            na.linspace(300, 400, axis="wavelength", num=5) * u.AA,
        ],
    )
    @pytest.mark.parametrize(
        argnames="direction_before",
        argvalues=[1],
    )
    @pytest.mark.parametrize(
        argnames="direction_after",
        argvalues=[1],
    )
    @pytest.mark.parametrize("n_before", [1])
    @pytest.mark.parametrize("n_after", [1.5])
    def test_transmissivity(
        self,
        a: optika.materials.profiles.AbstractInterfaceProfile,
        wavelength: u.Quantity | na.AbstractScalar,
        direction_before: float | na.AbstractScalar,
        direction_after: float | na.AbstractScalar,
        n_before: float | na.AbstractScalar,
        n_after: float | na.AbstractScalar,
    ):
        result = a.transmissivity(
            wavelength=wavelength,
            direction_before=direction_before,
            direction_after=direction_after,
            n_before=n_before,
            n_after=n_after,
        )
        assert na.unit_normalized(result).is_equivalent(u.dimensionless_unscaled)
        assert np.all(result >= 0)
        assert np.all(result <= 1)

    @pytest.mark.parametrize(
        argnames="wavelength",
        argvalues=[
            304 * u.AA,
            na.linspace(300, 400, axis="wavelength", num=5) * u.AA,
        ],
    )
    @pytest.mark.parametrize(
        argnames="direction",
        argvalues=[
            1,
            np.cos(na.linspace(-90, 90, axis="angle", num=5) * u.deg),
        ],
    )
    @pytest.mark.parametrize("n", [1.5])
    def test_reflectivity(
        self,
        a: optika.materials.profiles.AbstractInterfaceProfile,
        wavelength: u.Quantity | na.AbstractScalar,
        direction: float | na.AbstractScalar,
        n: float | na.AbstractScalar,
    ):
        result = a.reflectivity(
            wavelength=wavelength,
            direction=direction,
            n=n,
        )
        assert na.unit_normalized(result).is_equivalent(u.dimensionless_unscaled)
        assert np.all(result >= 0)
        assert np.all(result <= 1)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.profiles.ErfInterfaceProfile(width=5 * u.AA),
    ],
)
class TestErfInterfaceProfile(
    AbstractTestAbstractInterfaceProfile,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.profiles.ExponentialInterfaceProfile(
            width=5 * u.AA,
        ),
        optika.materials.profiles.ExponentialInterfaceProfile(
            width=na.linspace(5, 10, axis="width", num=6) * u.AA,
        ),
        optika.materials.profiles.ExponentialInterfaceProfile(
            width=na.NormalUncertainScalarArray(5, 1, num_distribution=11) * u.AA,
        ),
    ],
)
class TestExponentialInterfaceProfile(
    AbstractTestAbstractInterfaceProfile,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.profiles.LinearInterfaceProfile(
            width=5 * u.AA,
        ),
        optika.materials.profiles.LinearInterfaceProfile(
            width=na.linspace(5, 10, axis="width", num=6) * u.AA,
        ),
        optika.materials.profiles.LinearInterfaceProfile(
            width=na.NormalUncertainScalarArray(5, 1, num_distribution=11) * u.AA,
        ),
    ],
)
class TestLinearInterfaceProfile(
    AbstractTestAbstractInterfaceProfile,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.profiles.SinusoidalInterfaceProfile(
            width=5 * u.AA,
        ),
        optika.materials.profiles.SinusoidalInterfaceProfile(
            width=na.linspace(5, 10, axis="width", num=6) * u.AA,
        ),
        optika.materials.profiles.SinusoidalInterfaceProfile(
            width=na.NormalUncertainScalarArray(5, 1, num_distribution=11) * u.AA,
        ),
    ],
)
class TestSinusoidalInterfaceProfile(
    AbstractTestAbstractInterfaceProfile,
):
    pass



from typing import Sequence
import pytest
import pathlib
import numpy as np
import matplotlib
import astropy.units as u
import astropy.visualization
import named_arrays as na
import optika
from optika.materials import AbstractLayer, Layer
from . import test_materials

_wavelength = na.linspace(100, 200, axis="wavelength", num=4) * u.AA


@pytest.mark.parametrize("wavelength", [_wavelength])
@pytest.mark.parametrize(
    argnames="direction",
    argvalues=[
        1,
        np.cos(na.linspace(-1, 1, axis="angle", num=5)),
    ],
)
@pytest.mark.parametrize("n", [1])
@pytest.mark.parametrize(
    argnames="layers",
    argvalues=[
        Layer(
            chemical="SiC",
            thickness=10 * u.nm,
        ),
    ],
)
@pytest.mark.parametrize(
    argnames="substrate",
    argvalues=[
        None,
        Layer("SiO2"),
    ],
)
def test_multilayer_coefficients(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: float | na.AbstractScalar,
    n: float | na.AbstractScalar,
    layers: Sequence[AbstractLayer] | AbstractLayer,
    substrate: None | Layer,
):
    r, t = optika.materials.multilayer_coefficients(
        wavelength=wavelength,
        direction=direction,
        n=n,
        layers=layers,
        substrate=substrate,
    )

    assert np.all(np.real(r) >= -1)
    assert np.all(np.real(r) <= 1)

    assert np.all(np.imag(r) >= -1)
    assert np.all(np.imag(r) <= 1)


@pytest.mark.parametrize("wavelength", [_wavelength])
@pytest.mark.parametrize(
    argnames="direction",
    argvalues=[
        1,
        np.cos(na.linspace(-1, 1, axis="angle", num=5)),
    ],
)
@pytest.mark.parametrize("n", [1])
@pytest.mark.parametrize(
    argnames="layers",
    argvalues=[
        Layer(
            chemical="SiC",
            thickness=10 * u.nm,
        ),
    ],
)
@pytest.mark.parametrize(
    argnames="substrate",
    argvalues=[
        None,
        Layer("SiO2"),
    ],
)
def test_multilayer_efficiency(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: float | na.AbstractScalar,
    n: float | na.AbstractScalar,
    layers: Sequence[AbstractLayer] | AbstractLayer,
    substrate: None | Layer,
):
    reflected, transmitted = optika.materials.multilayer_efficiency(
        wavelength=wavelength,
        direction=direction,
        n=n,
        layers=layers,
        substrate=substrate,
    )

    assert np.all(reflected >= 0)
    assert np.all(reflected <= 1)
    assert np.all(transmitted >= 0)
    assert np.all(transmitted <= 1)
    assert np.all(reflected + transmitted <= 1)
    assert np.all(np.imag(reflected) == 0)
    assert np.all(np.imag(transmitted) == 0)


@pytest.mark.parametrize("index", [0])
@pytest.mark.parametrize("wavelength", [100 * u.AA, _wavelength])
@pytest.mark.parametrize(
    argnames="direction",
    argvalues=[
        1,
        np.cos(na.linspace(-1, 1, axis="angle", num=5)),
    ],
)
@pytest.mark.parametrize("n", [1])
@pytest.mark.parametrize(
    argnames="layers",
    argvalues=[
        Layer(
            chemical="SiC",
            thickness=10 * u.nm,
        ),
        [
            Layer(
                chemical="SiC",
                thickness=10 * u.nm,
            )
        ],
    ],
)
@pytest.mark.parametrize(
    argnames="substrate",
    argvalues=[
        None,
        Layer("SiO2"),
    ],
)
def test_layer_absorbance(
    index: int,
    wavelength: u.Quantity | na.AbstractScalar,
    direction: float | na.AbstractScalar,
    n: float | na.AbstractScalar,
    layers: None | Sequence[AbstractLayer] | optika.materials.AbstractLayer,
    substrate: None | Layer,
):
    result = optika.materials.layer_absorbance(
        index=index,
        wavelength=wavelength,
        direction=direction,
        n=n,
        layers=layers,
        substrate=substrate,
    )

    assert np.all(np.imag(result) == 0)
    assert np.all(result >= 0)
    assert np.all(result <= 1)

    reflectivity, transmissivity = optika.materials.multilayer_efficiency(
        wavelength=wavelength,
        direction=direction,
        n=n,
        layers=layers,
        substrate=substrate,
    )

    result_expected = 1 - reflectivity - transmissivity

    assert np.allclose(result, result_expected)


@pytest.mark.parametrize(
    argnames=[
        "file",
        "direction",
        "n",
        "layers",
        "substrate",
        "is_mirror",
    ],
    argvalues=[
        (
            pathlib.Path(__file__).parent / "_data/Si.txt",
            1,
            1,
            None,
            Layer("Si"),
            False,
        ),
        (
            pathlib.Path(__file__).parent / "_data/SiO2.txt",
            1,
            1,
            Layer("SiO2", thickness=50 * u.AA),
            Layer("Si"),
            False,
        ),
        (
            pathlib.Path(__file__).parent / "_data/SiO2_100A.txt",
            1,
            1,
            Layer("SiO2", thickness=100 * u.AA),
            Layer("Si"),
            False,
        ),
        (
            pathlib.Path(__file__).parent / "_data/SiC_Cr.txt",
            1,
            1,
            [
                Layer("SiC", thickness=25 * u.nm),
                Layer("Cr", thickness=5 * u.nm),
            ],
            Layer("SiO2"),
            True,
        ),
        pytest.param(
            pathlib.Path(__file__).parent / "_data/SiC_Cr_Rough.txt",
            1,
            1,
            [
                Layer(
                    chemical="SiC",
                    thickness=25 * u.nm,
                    interface=optika.materials.profiles.ErfInterfaceProfile(2 * u.nm),
                ),
                Layer(
                    chemical="Cr",
                    thickness=5 * u.nm,
                    interface=optika.materials.profiles.ErfInterfaceProfile(2 * u.nm),
                ),
            ],
            Layer(
                chemical="SiO2",
                interface=optika.materials.profiles.ErfInterfaceProfile(2 * u.nm),
            ),
            True,
            marks=pytest.mark.xfail(
                reason="IMD incorrectly uses the vacuum wavelength"
            ),
        ),
    ],
)
def test_multilayer_efficiency_vs_file(
    file: pathlib.Path,
    direction: float | na.AbstractScalar,
    n: float | na.AbstractScalar,
    layers: Sequence[AbstractLayer] | optika.materials.AbstractLayer,
    substrate: None | Layer,
    is_mirror: bool,
):
    skip_header = 0
    with open(file, "r") as f:
        for line in f:
            if line.startswith(";"):
                skip_header += 1
            else:
                break

    wavelength, efficiency_file = np.genfromtxt(
        fname=file,
        skip_header=skip_header,
        unpack=True,
    )
    wavelength = na.ScalarArray(wavelength, axes="wavelength") << u.AA
    efficiency_file = na.ScalarArray(efficiency_file, axes="wavelength")

    reflectivity, transmissivity = optika.materials.multilayer_efficiency(
        wavelength=wavelength,
        direction=direction,
        n=n,
        layers=layers,
        substrate=substrate,
    )

    if is_mirror:
        efficiency = reflectivity.average
    else:
        efficiency = transmissivity.average

    assert np.allclose(efficiency, efficiency_file, rtol=1e-4)


class AbstractTestAbstractMultilayerMaterial(
    test_materials.AbstractTestAbstractMaterial,
):
    def test_layers(self, a: optika.materials.AbstractMultilayerMaterial):
        result = a.layers
        if not isinstance(result, optika.materials.AbstractLayer):
            for layer in result:
                assert isinstance(layer, optika.materials.AbstractLayer)

    @pytest.mark.parametrize("ax", [None])
    def test_plot_layers(
        self,
        a: optika.materials.AbstractMultilayerMaterial,
        ax: None | matplotlib.axes.Axes,
    ):
        with astropy.visualization.quantity_support():
            result = a.plot_layers(
                ax=ax,
            )

        assert isinstance(result, list)
        for item in result:
            assert isinstance(item, matplotlib.patches.Rectangle)


class AbstractTestAbstractMultilayerFilm(
    AbstractTestAbstractMultilayerMaterial,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.MultilayerFilm(
            layers=optika.materials.LayerSequence(
                [
                    optika.materials.Layer("Al2O3", thickness=5 * u.nm),
                    optika.materials.Layer("Al", thickness=100 * u.nm),
                    optika.materials.Layer("Al2O3", thickness=5 * u.nm),
                ],
            ),
        ),
        optika.materials.MultilayerFilm(
            layers=optika.materials.PeriodicLayerSequence(
                [
                    optika.materials.Layer("Al2O3", thickness=5 * u.nm),
                    optika.materials.Layer("Al", thickness=100 * u.nm),
                ],
                num_periods=2,
            ),
        ),
    ],
)
class TestMultilayerFilm(
    AbstractTestAbstractMultilayerFilm,
):
    pass


class AbstractTestAbstractMultilayerMirror(
    AbstractTestAbstractMultilayerMaterial,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.MultilayerMirror(
            layers=[
                optika.materials.Layer("Al2O3", thickness=5 * u.nm),
                optika.materials.Layer("Al", thickness=100 * u.nm),
                optika.materials.Layer("Al2O3", thickness=5 * u.nm),
            ],
            substrate=optika.materials.Layer("SiO2", thickness=10 * u.mm),
        ),
        optika.materials.MultilayerMirror(
            layers=optika.materials.PeriodicLayerSequence(
                [
                    optika.materials.Layer("Al2O3", thickness=5 * u.nm),
                    optika.materials.Layer("Al", thickness=100 * u.nm),
                    optika.materials.Layer("Al2O3", thickness=5 * u.nm),
                ],
                num_periods=2,
            ),
            substrate=optika.materials.Layer("SiO2", thickness=10 * u.mm),
        ),
    ],
)
class TestMultilayerMirror(
    AbstractTestAbstractMultilayerMirror,
):
    pass



"""
A subpackage for simulating the interaction of light with transparent
or reflective materials.
"""

from ._snells_law import (
    snells_law_scalar,
    snells_law,
)
from . import profiles
from . import matrices
from ._layers import (
    AbstractLayer,
    Layer,
    AbstractLayerSequence,
    LayerSequence,
    PeriodicLayerSequence,
)
from ._materials import (
    AbstractMaterial,
    Vacuum,
    AbstractMirror,
    Mirror,
    MeasuredMirror,
)
from ._multilayers import (
    multilayer_coefficients,
    multilayer_efficiency,
    layer_absorbance,
    AbstractMultilayerMaterial,
    AbstractMultilayerFilm,
    MultilayerFilm,
    AbstractMultilayerMirror,
    MultilayerMirror,
)
from . import meshes
from ._thin_films import (
    AbstractThinFilmFilter,
    ThinFilmFilter,
)

__all__ = [
    "snells_law_scalar",
    "snells_law",
    "profiles",
    "matrices",
    "AbstractLayer",
    "Layer",
    "AbstractLayerSequence",
    "LayerSequence",
    "PeriodicLayerSequence",
    "AbstractMaterial",
    "Vacuum",
    "AbstractMirror",
    "Mirror",
    "MeasuredMirror",
    "multilayer_coefficients",
    "multilayer_efficiency",
    "layer_absorbance",
    "AbstractMultilayerMaterial",
    "AbstractMultilayerFilm",
    "MultilayerFilm",
    "AbstractMultilayerMirror",
    "MultilayerMirror",
    "meshes",
    "AbstractThinFilmFilter",
    "ThinFilmFilter",
]



"""
Image sensors used to measure the light intensity at the output of an optical
system.
"""

from ._materials import (
    charge_diffusion,
    mean_charge_capture,
    energy_bandgap,
    energy_electron_hole,
    quantum_yield_ideal,
    absorbance,
    charge_collection_efficiency,
    quantum_efficiency_effective,
    probability_measurement,
    electrons_measured,
    signal,
    AbstractDepletionModel,
    AbstractJanesickDepletionModel,
    E2VCCD64ThickDepletionModel,
    E2VCCD64ThinDepletionModel,
    AbstractImagingSensorMaterial,
    IdealImagingSensorMaterial,
    AbstractCCDMaterial,
    AbstractBackilluminatedCCDMaterial,
    AbstractStern1994BackilluminatedCCDMaterial,
    TektronixTK512CBMaterial,
    E2VCCD97Material,
    E2VCCD203Material,
)
from ._sensors import (
    AbstractImagingSensor,
    ImagingSensor,
    AbstractCCD,
)

__all__ = [
    "charge_diffusion",
    "mean_charge_capture",
    "energy_bandgap",
    "energy_electron_hole",
    "quantum_yield_ideal",
    "absorbance",
    "charge_collection_efficiency",
    "quantum_efficiency_effective",
    "probability_measurement",
    "electrons_measured",
    "signal",
    "AbstractDepletionModel",
    "AbstractJanesickDepletionModel",
    "E2VCCD64ThickDepletionModel",
    "E2VCCD64ThinDepletionModel",
    "AbstractImagingSensorMaterial",
    "IdealImagingSensorMaterial",
    "AbstractCCDMaterial",
    "AbstractBackilluminatedCCDMaterial",
    "AbstractStern1994BackilluminatedCCDMaterial",
    "TektronixTK512CBMaterial",
    "E2VCCD97Material",
    "E2VCCD203Material",
    "AbstractImagingSensor",
    "ImagingSensor",
    "AbstractCCD",
]



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
from named_arrays._vectors.cartesian.tests import test_vectors_cartesian
import optika

_num_y = test_vectors_cartesian._num_y


rays = [
    optika.rays.RayVectorArray(
        wavelength=500 * u.nm,
        position=na.Cartesian3dVectorArray(0, 1, 2) * u.mm,
        direction=na.Cartesian3dVectorArray(0, 0, 1),
        attenuation=0 / u.mm,
    ),
    optika.rays.RayVectorArray(
        wavelength=na.linspace(
            start=400 * u.nm,
            stop=500 * u.nm,
            axis="y",
            num=_num_y,
        ),
        position=na.Cartesian3dVectorLinearSpace(
            start=-10 * u.mm,
            stop=10 * u.mm,
            axis="y",
            num=_num_y,
        ).explicit,
        direction=na.Cartesian3dVectorArray(
            x=np.sin(na.linspace(-10, 10, axis="y", num=_num_y) * u.deg),
            y=0,
            z=np.cos(na.linspace(-10, 10, axis="y", num=_num_y) * u.deg),
        ),
        attenuation=na.linspace(0, 1, axis="y", num=_num_y) / u.mm,
        index_refraction=na.linspace(1, 1.5, axis="y", num=_num_y).explicit,
    ),
    optika.rays.RayVectorArray(
        wavelength=na.NormalUncertainScalarArray(600 * u.nm, width=5 * u.nm).explicit,
        position=na.Cartesian3dVectorArray(
            x=na.NormalUncertainScalarArray(3 * u.mm, width=1 * u.mm),
            y=na.NormalUncertainScalarArray(4 * u.mm, width=2 * u.mm),
            z=na.NormalUncertainScalarArray(5 * u.mm, width=3 * u.mm),
        ).explicit,
        direction=na.Cartesian3dVectorArray(0, 0, 1),
        attenuation=na.UniformUncertainScalarArray(0.5, width=0.25) / u.mm,
        index_refraction=na.UniformUncertainScalarArray(1.5, width=0.5).explicit,
    ),
]


def _items() -> list[na.AbstractArray | dict[str, int | slice | na.AbstractArray]]:
    return [
        dict(y=0),
        dict(y=slice(0, 1)),
        dict(y=na.ScalarArrayRange(0, 2, axis="y")),
    ]


class AbstractTestAbstractRayVectorArray(
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray
):
    def test_direction(self, array: optika.rays.AbstractRayVectorArray):
        assert isinstance(
            na.as_named_array(array.direction),
            na.AbstractCartesian3dVectorArray,
        )
        assert np.allclose(array.direction.length, 1)

    def test_intensity(self, array: optika.rays.AbstractRayVectorArray):
        assert isinstance(na.as_named_array(array.intensity), na.AbstractScalar)
        assert np.all(array.intensity >= 0)

    def test_attenuation(self, array: optika.rays.AbstractRayVectorArray):
        assert isinstance(na.as_named_array(array.attenuation), na.AbstractScalar)
        assert np.all(array.attenuation >= 0)
        assert na.unit_normalized(array.attenuation).is_equivalent(1 / u.m)

    def test_index_refraction(self, array: optika.rays.AbstractRayVectorArray):
        assert isinstance(na.as_named_array(array.index_refraction), na.AbstractScalar)

    def test_unvignetted(self, array: optika.rays.AbstractRayVectorArray):
        assert isinstance(na.as_named_array(array.unvignetted), na.AbstractScalar)

    @pytest.mark.xfail(raises=NotImplementedError)
    def test_matrix(self, array):
        return super().test_matrix(array=array)

    @pytest.mark.parametrize(
        argnames="item",
        argvalues=_items(),
    )
    def test__getitem__(
        self,
        array: na.AbstractPositionalVectorArray,
        item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray,
    ):
        super().test__getitem__(array=array, item=item)

    @pytest.mark.parametrize("array_2", rays)
    class TestUfuncBinary(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestUfuncBinary
    ):
        pass

    @pytest.mark.parametrize(
        argnames="array_2",
        argvalues=[
            na.Cartesian3dVectorArray(1, 2, 3) * u.mm,
        ],
    )
    class TestAddCartesian3dVector:
        def test_add_cartesian3d_vector(
            self,
            array: optika.rays.AbstractRayVectorArray,
            array_2: na.AbstractCartesian3dVectorArray,
        ):
            result = array + array_2
            assert result is not array
            assert np.all(result.position == array.position + array_2)
            assert result.wavelength is array.wavelength
            assert result.direction is array.direction
            assert result.index_refraction is array.index_refraction
            assert result.attenuation is array.attenuation
            assert result.intensity is array.intensity
            assert result.unvignetted is array.unvignetted

        def test_add_cartesian3d_vector_reversed(
            self,
            array: optika.rays.AbstractRayVectorArray,
            array_2: na.AbstractCartesian3dVectorArray,
        ):
            result = array_2 + array
            assert result is not array
            assert np.all(result.position == array_2 + array.position)
            assert result.wavelength is array.wavelength
            assert result.direction is array.direction
            assert result.index_refraction is array.index_refraction
            assert result.attenuation is array.attenuation
            assert result.intensity is array.intensity
            assert result.unvignetted is array.unvignetted

    @pytest.mark.parametrize("array_2", rays)
    class TestMatmul(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestMatmul
    ):
        pass

    @pytest.mark.parametrize(
        argnames="array_2",
        argvalues=[
            na.Cartesian3dXRotationMatrixArray(30 * u.deg),
        ],
    )
    class TestMatmulCartesian3dMatrix:
        def test_matmul_cartesian3d_matrix(
            self,
            array: optika.rays.AbstractRayVectorArray,
            array_2: na.AbstractCartesian3dVectorArray,
        ):
            result = array @ array_2
            assert result is not array
            assert np.all(result.position == array.position @ array_2)
            assert np.all(result.direction == array.direction @ array_2)
            assert result.wavelength is array.wavelength
            assert result.index_refraction is array.index_refraction
            assert result.attenuation is array.attenuation
            assert result.intensity is array.intensity
            assert result.unvignetted is array.unvignetted

        def test_matmul_cartesian3d_matrix_reversed(
            self,
            array: optika.rays.AbstractRayVectorArray,
            array_2: na.AbstractCartesian3dVectorArray,
        ):
            result = array_2 @ array
            assert result is not array
            assert np.all(result.position == array_2 @ array.position)
            assert np.all(result.direction == array_2 @ array.direction)
            assert result.wavelength is array.wavelength
            assert result.index_refraction is array.index_refraction
            assert result.attenuation is array.attenuation
            assert result.intensity is array.intensity
            assert result.unvignetted is array.unvignetted

    class TestArrayFunctions(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions
    ):
        @pytest.mark.parametrize("array_2", rays)
        class TestAsArrayLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestAsArrayLikeFunctions
        ):
            pass

        @pytest.mark.parametrize(
            argnames="where",
            argvalues=[
                np._NoValue,
            ],
        )
        class TestReductionFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestReductionFunctions,
        ):
            pass

        @pytest.mark.parametrize(
            argnames="q",
            argvalues=[
                25 * u.percent,
            ],
        )
        class TestPercentileLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestPercentileLikeFunctions,
        ):
            pass

    class TestNamedArrayFunctions(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions,
    ):
        @pytest.mark.skip
        class TestPltPlotLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions.TestPltPlotLikeFunctions,
        ):
            pass

        @pytest.mark.skip
        class TestJacobian(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions.TestJacobian,
        ):
            pass

        @pytest.mark.skip
        class TestOptimizeRoot(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions.TestOptimizeRoot,
        ):
            pass


@pytest.mark.parametrize("array", rays)
class TestRayVectorArray(
    AbstractTestAbstractRayVectorArray,
    test_vectors_cartesian.AbstractTestAbstractExplicitCartesianVectorArray,
):
    @pytest.mark.parametrize(
        argnames="item",
        argvalues=[
            dict(y=0),
            dict(y=slice(None)),
        ],
    )
    @pytest.mark.parametrize(
        argnames="value",
        argvalues=[
            optika.rays.RayVectorArray(
                wavelength=700 * u.nm,
                position=5 * u.mm,
                direction=na.Cartesian3dVectorArray(0, 1, 0),
            ),
        ],
    )
    def test__setitem__(
        self,
        array: na.ScalarArray,
        item: dict[str, int | slice | na.ScalarArray] | na.ScalarArray,
        value: float | na.ScalarArray,
    ):
        super().test__setitem__(array=array, item=item, value=value)



from __future__ import annotations
from typing import Any
import abc
import dataclasses
import numpy as np
import numpy.typing as npt
import matplotlib.axes
import astropy.units as u
import named_arrays as na

__all__ = [
    "Shaped",
    "Printable",
    "Plottable",
    "Transformable",
    "Translatable",
    "Pitchable",
    "Yawable",
    "Rollable",
]


@dataclasses.dataclass(repr=False)
class Shaped(abc.ABC):

    @property
    @abc.abstractmethod
    def shape(self) -> dict[str, int]:
        """
        The array shape of this object.
        """


@dataclasses.dataclass(repr=False)
class Printable:
    @classmethod
    def _val_to_string(
        cls,
        val: Any,
        pre: str,
        tab: str,
        field_str: str,
    ) -> str:
        if isinstance(val, Printable):
            val_str = val.to_string(prefix=f"{pre}{tab}")
        elif isinstance(val, na.AbstractArray):
            val_str = val.to_string(prefix=f"{pre}{tab}")
        elif isinstance(val, np.ndarray):
            val_str = np.array2string(
                a=val,
                separator=", ",
                prefix=field_str,
            )
            if isinstance(val, u.Quantity):
                val_str = f"{val_str} {val.unit}"
        elif isinstance(val, list):
            val_str = "[\n"
            for v in val:
                val_str += f"{pre}{tab}{tab}"
                val_str += cls._val_to_string(
                    val=v,
                    pre=f"{pre}{tab}",
                    tab=tab,
                    field_str=f"{pre}{tab}",
                )
                val_str += ",\n"
            val_str += f"{pre}{tab}]"
        else:
            val_str = repr(val)

        return val_str

    def to_string(
        self,
        prefix: None | str = None,
    ) -> str:
        """
        Public-facing version of the ``__repr__`` method that allows for
        defining a prefix string, which can be used to calculate how much
        whitespace to add to the beginning of each line of the result.

        Parameters
        ----------
        prefix
            an optional string, the length of which is used to calculate how
            much whitespace to add to the result.
        """

        fields = dataclasses.fields(self)

        delim_field = "\n"
        pre = " " * len(prefix) if prefix is not None else ""
        tab = " " * 4

        result_fields = ""
        for i, f in enumerate(fields):
            field_str = f"{pre}{tab}{f.name}="
            val = getattr(self, f.name)
            val_str = self._val_to_string(
                val=val,
                pre=pre,
                tab=tab,
                field_str=field_str,
            )
            field_str += val_str
            field_str += f",{delim_field}"
            result_fields += field_str

        if result_fields:
            result_fields = f"\n{result_fields}{pre}"

        result = f"{self.__class__.__qualname__}({result_fields})"

        return result

    def __repr__(self):
        return self.to_string()


@dataclasses.dataclass(eq=False, repr=False)
class Plottable(abc.ABC):
    @property
    @abc.abstractmethod
    def kwargs_plot(self) -> None | dict:
        """
        Extra keyword arguments that will be used in the call to
        :func:`named_arrays.plt.plot` within the :meth:`plot` method.
        """

    @abc.abstractmethod
    def plot(
        self,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray] = None,
        transformation: None | na.transformations.AbstractTransformation = None,
        components: None | tuple[str, ...] = None,
        **kwargs,
    ) -> na.AbstractScalar | dict[str, na.AbstractScalar]:
        """
        Plot the selected components onto the given axes.

        Parameters
        ----------
        ax
            The matplotlib axes to plot onto
        transformation
            Any extra transformations to apply to the coordinate system before
            plotting
        components
            Which 3d components to plot, helpful if plotting in 2d.
        kwargs
            Additional keyword arguments that will be passed along to
            :func:`named_arrays.plt.plot()`
        """


@dataclasses.dataclass(eq=False, repr=False)
class Transformable(abc.ABC):
    @property
    @abc.abstractmethod
    def transformation(self) -> None | na.transformations.AbstractTransformation:
        """
        the coordinate transformation between the global coordinate system
        and this object's local coordinate system
        """
        return na.transformations.IdentityTransformation()


@dataclasses.dataclass(eq=False, repr=False)
class Translatable(
    Transformable,
):
    @property
    @abc.abstractmethod
    def translation(self) -> u.Quantity | na.AbstractScalar | na.AbstractVectorArray:
        """translate the coordinate system"""

    @property
    def transformation(self) -> na.transformations.AbstractTransformation:
        translation = na.asanyarray(self.translation, like=na.Cartesian3dVectorArray())
        return super().transformation @ na.transformations.Translation(translation)


@dataclasses.dataclass(eq=False, repr=False)
class Pitchable(
    Transformable,
):
    @property
    @abc.abstractmethod
    def pitch(self) -> u.Quantity | na.ScalarLike:
        """pitch angle of this object"""

    @property
    def transformation(self) -> na.transformations.AbstractTransformation:
        return super().transformation @ na.transformations.Cartesian3dRotationX(
            angle=self.pitch
        )


@dataclasses.dataclass(eq=False, repr=False)
class Yawable(
    Transformable,
):
    @property
    @abc.abstractmethod
    def yaw(self) -> u.Quantity | na.ScalarLike:
        """yaw angle of this object"""

    @property
    def transformation(self) -> na.transformations.AbstractTransformation:
        return super().transformation @ na.transformations.Cartesian3dRotationY(
            angle=self.yaw,
        )


@dataclasses.dataclass(eq=False, repr=False)
class Rollable(
    Transformable,
):
    @property
    @abc.abstractmethod
    def roll(self) -> u.Quantity | na.ScalarLike:
        """roll angle of this object"""

    @property
    def transformation(self) -> na.transformations.AbstractTransformation:
        return super().transformation @ na.transformations.Cartesian3dRotationZ(
            angle=self.roll
        )



import abc
import functools
import dataclasses
import numpy as np
import scipy.optimize
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "AbstractDepletionModel",
    "AbstractJanesickDepletionModel",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractDepletionModel(
    optika.mixins.Printable,
    optika.mixins.Shaped,
):
    """
    An arbitrary model of the depletion region of a semiconducting imaging
    sensor.
    """

    @property
    @abc.abstractmethod
    def thickness(self) -> u.Quantity:
        """The thickness of the depletion region."""


@dataclasses.dataclass(eq=False, repr=False)
class AbstractJanesickDepletionModel(
    AbstractDepletionModel,
):
    """
    A depletion model that relies on :cite:t:`Janesick2001`'s model of
    charge diffusion to measure the thickness of the field-free region
    of an imaging sensor and infer the thickness of the depletion region.
    """

    @property
    @abc.abstractmethod
    def chemical_substrate(self) -> optika.chemicals.AbstractChemical:
        """
        The optical properties of the substrate material.
        """

    @property
    @abc.abstractmethod
    def thickness_substrate(self) -> u.Quantity:
        """
        The thickness of the light-sensitive layer of the imaging sensor.
        """

    @property
    @abc.abstractmethod
    def width_pixel(self) -> u.Quantity:
        """The size of a pixel."""

    def mean_charge_capture(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ) -> na.AbstractScalar:
        """
        The mean charge capture of this sensor for the given wavelength
        computed using :func:`optika.sensors.mean_charge_capture`

        Parameters
        ----------
        wavelength
            The wavelengths at which to evaluate the mean charge capture.
        """
        return optika.sensors.mean_charge_capture(
            width_diffusion=optika.sensors.charge_diffusion(
                absorption=self.chemical_substrate.absorption(wavelength),
                thickness_substrate=self.thickness_substrate,
                thickness_depletion=self.thickness,
            ),
            width_pixel=self.width_pixel,
        )

    @property
    @abc.abstractmethod
    def mean_charge_capture_measured(self) -> na.FunctionArray:
        """
        The measured mean charge capture that will be fit by the function
        :func:`optika.sensors.mean_charge_capture`.
        """

    @functools.cached_property
    def thickness(self) -> u.Quantity:

        thickness_substrate = self.thickness_substrate
        width_pixel = self.width_pixel

        mcc_measured = self.mean_charge_capture_measured

        absorption = self.chemical_substrate.absorption(mcc_measured.inputs)

        unit = u.um

        def objective(thickness_depletion: float) -> float:

            width_diffusion = optika.sensors.charge_diffusion(
                absorption=absorption,
                thickness_substrate=thickness_substrate,
                thickness_depletion=thickness_depletion * unit,
            )

            mcc = optika.sensors.mean_charge_capture(
                width_diffusion=width_diffusion,
                width_pixel=width_pixel,
            )

            diff = mcc_measured.outputs - mcc

            result = np.sqrt(np.mean(np.square(diff))).ndarray

            return result

        fit = scipy.optimize.minimize_scalar(
            fun=objective,
            bounds=(
                0,
                thickness_substrate.to_value(unit),
            ),
        )

        return fit.x * unit



"""
A collection of apertures that can be applied to optical surfaces to block light.
"""

import abc
import dataclasses
import numpy as np
import numpy.typing as npt
import matplotlib.axes
import matplotlib.lines
import matplotlib.pyplot as plt
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "AbstractAperture",
    "CircularAperture",
    "CircularSectorAperture",
    "EllipticalAperture",
    "AbstractPolygonalAperture",
    "PolygonalAperture",
    "AbstractRegularPolygonalAperture",
    "AbstractOctagonalAperture",
    "OctagonalAperture",
    "AbstractIsoscelesTrapezoidalAperture",
    "IsoscelesTrapezoidalAperture",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractAperture(
    optika.mixins.Printable,
    optika.mixins.Plottable,
    optika.mixins.Transformable,
    optika.mixins.Shaped,
):
    @property
    @abc.abstractmethod
    def samples_wire(self):
        """
        default number of samples used for :meth:`wire`
        """

    @property
    @abc.abstractmethod
    def active(self):
        """
        flag controlling whether the aperture can clip rays during a raytrace
        """

    @property
    @abc.abstractmethod
    def inverted(self):
        """
        flag controlling whether the interior or the exterior of the aperture
        allows light to pass through.
        If :obj:`True`, the interior of the aperture allows light to pass
        through.
        If :obj:`False`, the exterior of the aperture allows light to pass
        through.
        """

    @abc.abstractmethod
    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        Check if a given point is inside the aperture.

        Parameters
        ----------
        position
            the points to check
        """

    def clip_rays(self, rays: optika.rays.RayVectorArray):
        unit = na.unit_normalized(self.bound_lower)
        if unit.is_equivalent(u.mm):
            mask = self(rays.position)
        elif unit.is_equivalent(u.dimensionless_unscaled):
            mask = self(rays.direction)
        else:
            raise ValueError(f"aperture with unit {unit} is not supported")
        rays = rays.copy_shallow()
        rays.unvignetted = rays.unvignetted & mask
        return rays

    @property
    @abc.abstractmethod
    def bound_lower(self) -> na.AbstractCartesian3dVectorArray:
        """
        The lower-left corner of the aperture's rectangular footprint
        """

    @property
    @abc.abstractmethod
    def bound_upper(self) -> na.AbstractCartesian3dVectorArray:
        """
        The upper-right corner of the aperture's rectangular footprint
        """

    @property
    @abc.abstractmethod
    def vertices(self) -> None | na.AbstractCartesian3dVectorArray:
        """
        The vertices of the polygon representing this aperture
        """

    @abc.abstractmethod
    def wire(self, num: None | int = None) -> na.AbstractCartesian3dVectorArray:
        """
        The sequence of points representing this aperture

        Parameters
        ----------
        num
            The total number of samples that will be used to represent this
            wire.
        """

    def plot(
        self,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray] = None,
        transformation: None | na.transformations.AbstractTransformation = None,
        components: None | tuple[str, ...] = None,
        sag: None | optika.sags.AbstractSag = None,
        **kwargs,
    ) -> None | na.ScalarArray[npt.NDArray[None | matplotlib.lines.Line2D]]:
        if ax is None:
            ax = plt.gca()
        ax = na.as_named_array(ax)

        wire = self.wire().explicit

        if not wire.length.unit_normalized.is_equivalent(u.mm):
            return None

        if sag is not None:
            wire.z = sag(wire)

        kwargs_plot = self.kwargs_plot
        if kwargs_plot is None:
            kwargs_plot = dict()

        kwargs = kwargs_plot | kwargs

        return na.plt.plot(
            wire,
            ax=ax,
            axis="wire",
            transformation=transformation,
            components=components,
            **kwargs,
        )


@dataclasses.dataclass(eq=False, repr=False)
class CircularAperture(
    AbstractAperture,
):
    """
    A circular aperture or obscuration

    Examples
    --------

    Plot a single circular aperture

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import numpy as np
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        aperture = optika.apertures.CircularAperture(50 * u.mm)

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(components=("x", "y"), color="black")

    |

    Plot an array of circular apertures, similar to the configuration of the
    `Giant Magellan Telescope <https://en.wikipedia.org/wiki/Giant_Magellan_Telescope>`_.

    .. jupyter-execute::

        diameter = 8.417 * u.m
        radius = diameter / 2

        angle = na.linspace(0, 360, axis="segment", num=6, endpoint=False) * u.deg

        displacement = na.Cartesian3dVectorArray(
            x=diameter * np.cos(angle),
            y=diameter * np.sin(angle),
        )
        displacement = np.concatenate([
            na.Cartesian3dVectorArray().add_axes("segment") * u.mm,
            displacement
        ], axis="segment")

        aperture = optika.apertures.CircularAperture(
            radius=radius,
            transformation=na.transformations.Translation(displacement),
        )

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(components=("x", "y"), color="black")
    """

    radius: u.Quantity | na.AbstractScalar = 0 * u.mm
    """the radius of the aperture"""

    samples_wire: int = 101
    active: bool | na.AbstractScalar = True
    inverted: bool | na.AbstractScalar = False
    transformation: None | na.transformations.AbstractTransformation = None
    kwargs_plot: None | dict = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.radius),
            optika.shape(self.active),
            optika.shape(self.inverted),
            optika.shape(self.transformation),
        )

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        radius = self.radius
        active = self.active
        inverted = self.inverted
        if self.transformation is not None:
            position = self.transformation.inverse(position)

        shape = na.shape_broadcasted(radius, active, inverted, position)

        radius = na.broadcast_to(radius, shape)
        active = na.broadcast_to(active, shape)
        inverted = na.broadcast_to(inverted, shape)
        position = na.broadcast_to(position, shape)

        mask = position.xy.length <= radius

        mask[inverted] = ~mask[inverted]
        mask[~active] = True

        return mask

    @property
    def bound_lower(self) -> na.Cartesian3dVectorArray:
        unit = na.unit(self.radius)
        result = na.Cartesian3dVectorArray()
        if unit is not None:
            result = result * unit
        if self.transformation is not None:
            result = self.transformation(result)
        result.x = result.x - self.radius
        result.y = result.y - self.radius
        return result

    @property
    def bound_upper(self) -> na.Cartesian3dVectorArray:
        unit = na.unit(self.radius)
        result = na.Cartesian3dVectorArray()
        if unit is not None:
            result = result * unit
        if self.transformation is not None:
            result = self.transformation(result)
        result.x = result.x + self.radius
        result.y = result.y + self.radius
        return result

    @property
    def vertices(self) -> None:
        return None

    def wire(self, num: None | int = None) -> na.Cartesian3dVectorArray:
        if num is None:
            num = self.samples_wire
        az = na.linspace(
            start=0 * u.deg,
            stop=360 * u.deg,
            axis="wire",
            num=num,
        )
        unit_radius = na.unit(self.radius)
        result = na.Cartesian3dVectorArray(
            x=self.radius * np.cos(az),
            y=self.radius * np.sin(az),
            z=0 * unit_radius if unit_radius is not None else 0,
        )
        if self.transformation is not None:
            result = self.transformation(result)
        return result


@dataclasses.dataclass(eq=False, repr=False)
class CircularSectorAperture(
    AbstractAperture,
):
    """
    A `circular sector <https://en.wikipedia.org/wiki/Circular_sector>`_
    aperture.

    Examples
    --------

    Plot a single circular aperture sector

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Define a circular aperture sector
        aperture = optika.apertures.CircularSectorAperture(
            radius=50 * u.mm,
            angle_start=-11 * u.deg,
            angle_stop=40 * u.deg,
        )

        # Define points to sample the aperture with
        points = na.Cartesian3dVectorLinearSpace(
            start=aperture.bound_lower,
            stop=aperture.bound_upper,
            axis=na.Cartesian3dVectorArray("x", "y", "z"),
            num=na.Cartesian3dVectorArray(11, 11, 1),
        )

        # Compute which points are inside the aperture
        where = aperture(points)

        # Plot the circular aperture sector
        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(components=("x", "y"), color="black")
            na.plt.scatter(
                points.x,
                points.y,
                c=where.astype(float)
            )
    """

    radius: u.Quantity | na.AbstractScalar = 0 * u.mm
    """
    The radius of the cirucular sector.
    """

    angle_start: u.Quantity | na.AbstractScalar = 0 * u.deg
    r"""
    The starting angle of the circular sector.
    Must be between :math:`-2 \pi` and :math:`+2 \pi` radians.
    """

    angle_stop: u.Quantity | na.AbstractScalar = 180 * u.deg
    """
    The ending angle of the circular sector.
    Must be between :math:`-2 \pi` and :math:`+2 \pi` radians and 
    counterclockwise from `angle_start`.
    """

    samples_wire: int = 101
    active: bool | na.AbstractScalar = True
    inverted: bool | na.AbstractScalar = False
    transformation: None | na.transformations.AbstractTransformation = None
    kwargs_plot: None | dict = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.radius),
            optika.shape(self.angle_start),
            optika.shape(self.angle_stop),
            optika.shape(self.active),
            optika.shape(self.inverted),
            optika.shape(self.transformation),
        )

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        radius = self.radius
        angle_start = self.angle_start
        angle_stop = self.angle_stop
        active = self.active
        inverted = self.inverted
        if self.transformation is not None:
            position = self.transformation.inverse(position)

        shape = na.shape_broadcasted(
            radius,
            angle_start,
            angle_stop,
            active,
            inverted,
            position,
        )

        radius = na.broadcast_to(radius, shape)
        angle_start = na.broadcast_to(angle_start, shape)
        angle_stop = na.broadcast_to(angle_stop, shape)
        active = na.broadcast_to(active, shape)
        inverted = na.broadcast_to(inverted, shape)
        position = na.broadcast_to(position, shape)

        mask_radius = position.xy.length <= radius

        angle = np.arctan2(position.y, position.x)
        angle_positive = angle % (+2 * np.pi * u.rad)
        angle_negative = angle % (-2 * np.pi * u.rad)
        mask_positive = (angle_start < angle_positive) & (angle_positive < angle_stop)
        mask_negative = (angle_start < angle_negative) & (angle_negative < angle_stop)
        mask_angle = mask_positive | mask_negative

        mask = mask_radius & mask_angle

        mask[inverted] = ~mask[inverted]
        mask[~active] = True

        return mask

    @property
    def bound_lower(self) -> na.Cartesian3dVectorArray:
        return self.wire().min()

    @property
    def bound_upper(self) -> na.Cartesian3dVectorArray:
        return self.wire().max()

    @property
    def vertices(self) -> None:
        return None

    def wire(self, num: None | int = None) -> na.Cartesian3dVectorArray:
        if num is None:
            num = self.samples_wire
        az = na.linspace(
            start=self.angle_start,
            stop=self.angle_stop,
            axis="wire",
            num=num - 2,
        )
        unit_radius = na.unit(self.radius)
        result = na.Cartesian3dVectorArray(
            x=self.radius * np.cos(az),
            y=self.radius * np.sin(az),
            z=0 * unit_radius if unit_radius is not None else 0,
        )

        vertex = na.Cartesian3dVectorArray().add_axes("wire")
        vertex = vertex * unit_radius if unit_radius is not None else vertex
        result = np.concatenate(
            [
                vertex,
                result,
                vertex,
            ],
            axis="wire",
        )
        if self.transformation is not None:
            result = self.transformation(result)
        return result


@dataclasses.dataclass(eq=False, repr=False)
class EllipticalAperture(
    AbstractAperture,
):
    """
    An elliptical aperture or obscuration

    Examples
    --------

    Plot a single elliptical aperture

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import numpy as np
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        aperture = optika.apertures.EllipticalAperture(
            na.Cartesian2dVectorArray(100, 50) * u.mm,
        )

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(components=("x", "y"), color="black")
    """

    radius: na.AbstractCartesian2dVectorArray = 0 * u.mm
    """the semi major/minor axes of the elliptical aperture."""

    samples_wire: int = 101
    """default number of samples used for :meth:`wire`"""

    active: bool | na.AbstractScalar = True
    """flag controlling whether the aperture can clip rays during a raytrace"""

    inverted: bool | na.AbstractScalar = False
    """
    flag controlling whether the interior or the exterior of the aperture
    allows light to pass through.
    If :obj:`True`, the interior of the aperture allows light to pass
    through.
    If :obj:`False`, the exterior of the aperture allows light to pass
    through.
    """

    transformation: None | na.transformations.AbstractTransformation = None
    """
    the coordinate transformation between the global coordinate system
    and this object's local coordinate system
    """

    kwargs_plot: None | dict = None
    """
    Extra keyword arguments that will be used in the call to
    :func:`named_arrays.plt.plot` within the :meth:`plot` method.
    """

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.radius),
            optika.shape(self.active),
            optika.shape(self.inverted),
            optika.shape(self.transformation),
        )

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:

        radius = self.radius
        active = self.active
        inverted = self.inverted
        if self.transformation is not None:
            position = self.transformation.inverse(position)

        shape = na.shape_broadcasted(radius, active, inverted, position)

        radius = na.broadcast_to(radius, shape)
        active = na.broadcast_to(active, shape)
        inverted = na.broadcast_to(inverted, shape)
        position = na.broadcast_to(position, shape)

        mask = np.square(position.x / radius.x) + np.square(position.y / radius.y) <= 1

        mask[inverted] = ~mask[inverted]
        mask[~active] = True

        return mask

    @property
    def bound_lower(self) -> na.Cartesian3dVectorArray:
        unit = na.unit(self.radius)
        result = na.Cartesian3dVectorArray()
        if unit is not None:
            result = result * unit
        if self.transformation is not None:
            result = self.transformation(result)
        result.x = result.x - self.radius.x
        result.y = result.y - self.radius.y
        return result

    @property
    def bound_upper(self) -> na.Cartesian3dVectorArray:
        unit = na.unit(self.radius)
        result = na.Cartesian3dVectorArray()
        if unit is not None:
            result = result * unit
        if self.transformation is not None:
            result = self.transformation(result)
        result.x = result.x + self.radius.x
        result.y = result.y + self.radius.y
        return result

    @property
    def vertices(self) -> None:
        return None

    def wire(self, num: None | int = None) -> na.Cartesian3dVectorArray:
        if num is None:
            num = self.samples_wire
        az = na.linspace(
            start=0 * u.deg,
            stop=360 * u.deg,
            axis="wire",
            num=num,
        )
        unit_radius = na.unit(self.radius)
        result = na.Cartesian3dVectorArray(
            x=self.radius.x * np.cos(az),
            y=self.radius.y * np.sin(az),
            z=0 * unit_radius if unit_radius is not None else 0,
        )
        if self.transformation is not None:
            result = self.transformation(result)
        return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractPolygonalAperture(
    AbstractAperture,
):
    """
    Base class for any type of polygonal aperture
    """

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        vertices = self.vertices
        active = self.active
        inverted = self.inverted
        if self.transformation is not None:
            position = self.transformation.inverse(position)

        shape = na.shape_broadcasted(
            vertices[dict(vertex=0)],
            active,
            inverted,
            position,
        )

        active = na.broadcast_to(active, shape)
        inverted = na.broadcast_to(inverted, shape)
        position = na.broadcast_to(position, shape)

        result = False
        for v in range(vertices.shape["vertex"]):
            vert_j = na.broadcast_to(vertices[dict(vertex=v - 1)], shape)
            vert_i = na.broadcast_to(vertices[dict(vertex=v)], shape)
            slope = (vert_j.y - vert_i.y) / (vert_j.x - vert_i.x)
            condition_1 = (vert_i.y > position.y) != (vert_j.y > position.y)
            condition_2 = position.x < ((position.y - vert_i.y) / slope + vert_i.x)
            result = result ^ (condition_1 & condition_2)

        result[inverted] = ~result[inverted]
        result[~active] = True

        return result

    @property
    def bound_lower(self) -> na.AbstractCartesian3dVectorArray:
        return self.vertices.min(axis="vertex")

    @property
    def bound_upper(self) -> na.AbstractCartesian3dVectorArray:
        return self.vertices.max(axis="vertex")

    def wire(self, num: None | int = None) -> na.Cartesian3dVectorArray:
        if num is None:
            num = self.samples_wire
        vertices = self.vertices.broadcasted
        num_vertices = vertices.shape["vertex"]
        num_sides = num_vertices
        num_per_side = num / num_sides
        index_right = na.arange(0, num_vertices, axis="vertex") + 1
        index_right = index_right % num_vertices
        index_right = dict(vertex=index_right)
        vertices_left = vertices
        vertices_right = vertices[index_right]
        wire = []
        num_cumulative = 0
        for v in range(num_vertices):
            num_v = int((v + 1) * num_per_side - num_cumulative)
            num_cumulative += num_v

            if num_cumulative == num:
                endpoint = True
            else:
                endpoint = False

            t = na.linspace(
                start=0,
                stop=1,
                axis="wire",
                num=num_v,
                endpoint=endpoint,
            )
            vertex_left = vertices_left[dict(vertex=v)]
            vertex_right = vertices_right[dict(vertex=v)]
            diff = vertex_right - vertex_left
            wire_v = vertex_left + diff * t
            wire.append(wire_v)

        wire = na.concatenate(wire, axis="wire")

        if self.transformation is not None:
            wire = self.transformation(wire)

        return wire


@dataclasses.dataclass(eq=False, repr=False)
class PolygonalAperture(
    AbstractPolygonalAperture,
):
    vertices: na.Cartesian3dVectorArray = 0 * u.mm
    samples_wire: int = 101
    active: bool | na.AbstractScalar = True
    inverted: bool | na.AbstractScalar = False
    transformation: None | na.transformations.AbstractTransformation = None
    kwargs_plot: None | dict = None

    @property
    def shape(self) -> dict[str, int]:
        shape = optika.shape(self.vertices)
        shape.pop("vertex")
        return na.broadcast_shapes(
            shape,
            optika.shape(self.active),
            optika.shape(self.inverted),
            optika.shape(self.transformation),
        )


@dataclasses.dataclass(eq=False, repr=False)
class RectangularAperture(
    AbstractPolygonalAperture,
):
    """
    A rectangular aperture or obscuration

    Examples
    --------

    Create a square aperture by setting :attr:`half_width` to a scalar value.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        half_width = 50 * u.mm

        aperture = optika.apertures.RectangularAperture(half_width)

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(components=("x", "y"), color="black")

    |

    Create a rectangular aperture by setting :attr:`half_width` to an
    instance of :class:`named_arrays.AbstractCartesian2dVectorArray`.

    .. jupyter-execute::

        half_width = na.Cartesian2dVectorArray(100, 50) * u.mm

        aperture = optika.apertures.RectangularAperture(half_width)

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(color="black")

    |

    Create a grid of rectangular apertures using the :attr:`transformation`
    parameter and the :class:`optika.transforms.Translation` transformation.

    .. jupyter-execute::

        pitch = 2 * half_width + 10 * u.mm

        displacement = na.Cartesian3dVectorArray(
            x=pitch.x * na.arange(0, 3, axis="aperture_x"),
            y=pitch.y * na.arange(0, 2, axis="aperture_y"),
        )

        aperture = optika.apertures.RectangularAperture(
            half_width=half_width,
            transformation=na.transformations.Translation(displacement),
        )

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(components=("x", "y"), color="black")
    """

    half_width: u.Quantity | na.AbstractScalar | na.Cartesian2dVectorArray = 0 * u.mm
    """distance from the origin to a perpendicular edge"""

    samples_wire: int = 101
    active: bool | na.AbstractScalar = True
    inverted: bool | na.AbstractScalar = False
    transformation: None | na.transformations.AbstractTransformation = None
    kwargs_plot: None | dict = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.half_width),
            optika.shape(self.active),
            optika.shape(self.inverted),
            optika.shape(self.transformation),
        )

    def __call__(
        self,
        position: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        bound_lower = self.bound_lower
        bound_upper = self.bound_upper
        active = self.active
        inverted = self.inverted
        if self.transformation is not None:
            position = self.transformation.inverse(position)

        shape = na.shape_broadcasted(
            bound_lower, bound_upper, active, inverted, position
        )

        bound_lower = na.broadcast_to(bound_lower, shape)
        bound_upper = na.broadcast_to(bound_upper, shape)
        active = na.broadcast_to(active, shape)
        inverted = na.broadcast_to(inverted, shape)
        position = na.broadcast_to(position, shape)

        mask = (bound_lower <= position) & (position <= bound_upper)
        mask = mask.x & mask.y

        mask[inverted] = ~mask[inverted]
        mask[~active] = True

        return mask

    @property
    def vertices(self):
        half_width = na.asanyarray(self.half_width, like=na.Cartesian2dVectorArray())
        r = np.sqrt(2)
        az = na.linspace(0, 360, axis="vertex", num=4, endpoint=False) * u.deg
        az = az + 45 * u.deg
        result = na.Cartesian3dVectorArray(
            x=r * np.cos(az).value,
            y=r * np.sin(az).value,
            z=0,
        )
        result.x = result.x * half_width.x
        result.y = result.y * half_width.y
        unit = na.unit(half_width.x)
        if unit is not None:
            result.z = result.z * unit
        return result


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRegularPolygonalAperture(
    AbstractPolygonalAperture,
):
    @property
    @abc.abstractmethod
    def radius(self) -> na.ScalarLike:
        """
        the radial distance from the origin to each vertex
        """

    @property
    @abc.abstractmethod
    def num_vertices(self) -> int:
        """
        Number of vertices in this polygon
        """

    @property
    def vertices(self) -> na.AbstractCartesian3dVectorArray:
        radius = self.radius
        unit = na.unit(radius)
        angle = na.linspace(
            start=0 * u.deg,
            stop=360 * u.deg,
            axis="vertex",
            num=self.num_vertices,
            endpoint=False,
        )
        result = na.Cartesian3dVectorArray(
            x=radius * np.cos(angle).value,
            y=radius * np.sin(angle).value,
            z=0,
        )
        if unit is not None:
            result.z = result.z * unit
        return result


@dataclasses.dataclass(eq=False, repr=False)
class RegularPolygonalAperture(
    AbstractRegularPolygonalAperture,
):
    radius: float | u.Quantity | na.AbstractScalar = 0 * u.mm
    num_vertices: int = 0
    samples_wire: int = 101
    active: bool | na.AbstractScalar = True
    inverted: bool | na.AbstractScalar = False
    transformation: None | na.transformations.AbstractTransformation = None
    kwargs_plot: None | dict = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.radius),
            optika.shape(self.active),
            optika.shape(self.inverted),
            optika.shape(self.transformation),
        )


@dataclasses.dataclass(eq=False, repr=False)
class AbstractOctagonalAperture(
    AbstractRegularPolygonalAperture,
):
    @property
    def num_vertices(self) -> int:
        return 8


@dataclasses.dataclass(eq=False, repr=False)
class OctagonalAperture(
    AbstractOctagonalAperture,
):
    radius: float | u.Quantity | na.AbstractScalar = 0 * u.mm
    samples_wire: int = 101
    active: bool | na.AbstractScalar = True
    inverted: bool | na.AbstractScalar = False
    transformation: None | na.transformations.AbstractTransformation = None
    kwargs_plot: None | dict = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.radius),
            optika.shape(self.active),
            optika.shape(self.inverted),
            optika.shape(self.transformation),
        )


@dataclasses.dataclass(eq=False, repr=False)
class AbstractIsoscelesTrapezoidalAperture(
    AbstractPolygonalAperture,
):
    @property
    @abc.abstractmethod
    def x_left(self) -> na.ScalarLike:
        """:math:`x` coordinate of the left base of the trapezoid"""

    @property
    @abc.abstractmethod
    def x_right(self) -> na.ScalarLike:
        """:math:`x` coordinate of the right base of the trapezoid"""

    @property
    @abc.abstractmethod
    def angle(self) -> na.ScalarLike:
        """angle between the two legs of the trapezoid"""

    @property
    def vertices(self) -> na.Cartesian3dVectorArray:
        x_left = self.x_left
        x_right = self.x_right
        angle = self.angle

        m = np.tan(angle / 2)
        left = na.Cartesian3dVectorArray(
            x=x_left,
            y=m * x_left,
            z=0 * x_left,
        )
        right = na.Cartesian3dVectorArray(
            x=x_right,
            y=m * x_right,
            z=0 * x_right,
        )

        upper = na.stack([left, right], axis="vertex")

        lower = upper[dict(vertex=slice(None, None, -1))]
        lower = lower * na.Cartesian3dVectorArray(1, -1, 1)

        result = na.concatenate([upper, lower], axis="vertex")

        return result


@dataclasses.dataclass(eq=False, repr=False)
class IsoscelesTrapezoidalAperture(
    AbstractIsoscelesTrapezoidalAperture,
):
    """
    This aperture is useful if you want to break a circular aperture up
    into different sectors.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        num_sectors = 8

        roll = na.linspace(0, 360, axis="roll", num=num_sectors, endpoint=False) * u.deg

        aperture = optika.apertures.IsoscelesTrapezoidalAperture(
            x_left=10 * u.mm,
            x_right=40 * u.mm,
            angle=(360 * u.deg) / num_sectors,
            transformation=na.transformations.TransformationList([
                na.transformations.Cartesian3dTranslation(x=5 * u.mm),
                na.transformations.Cartesian3dRotationZ(roll),
            ])
        )

        with astropy.visualization.quantity_support():
            plt.figure()
            plt.gca().set_aspect("equal")
            aperture.plot(components=("x", "y"), color="black")
    """

    x_left: na.ScalarLike = 0 * u.mm
    x_right: na.ScalarLike = 0 * u.mm
    angle: na.ScalarLike = 0 * u.deg
    samples_wire: int = 101
    active: bool | na.AbstractScalar = True
    inverted: bool | na.AbstractScalar = False
    transformation: None | na.transformations.AbstractTransformation = None
    kwargs_plot: None | dict = None

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.x_left),
            optika.shape(self.x_right),
            optika.shape(self.angle),
            optika.shape(self.active),
            optika.shape(self.inverted),
            optika.shape(self.transformation),
        )



import abc
import dataclasses
import astropy.units as u
import named_arrays as na
import optika

__all__ = [
    "AbstractRoughnessParameters",
    "RoughnessParameters",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRoughnessParameters(
    optika.mixins.Printable,
    optika.mixins.Shaped,
):
    """collection of parameters used to compute the roughness of an optical surface"""

    @property
    @abc.abstractmethod
    def period_min(self) -> na.ScalarLike:
        """
        minimum period to consider when calculating roughness
        """

    @property
    @abc.abstractmethod
    def period_max(self) -> na.ScalarLike:
        """maximum period to consider when calculating roughness"""


@dataclasses.dataclass(eq=False, repr=False)
class RoughnessParameters(
    AbstractRoughnessParameters,
):
    period_min: na.ScalarLike = 0 * u.mm
    period_max: na.ScalarLike = 0 * u.mm

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            optika.shape(self.period_min),
            optika.shape(self.period_max),
        )



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from optika._tests import test_mixins


class AbstractTestAbstractSlopeErrorParameters(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestShaped,
):
    def test_step_size(self, a: optika.metrology.AbstractSlopeErrorParameters):
        assert np.issubdtype(na.get_dtype(a.step_size), float)
        assert na.unit_normalized(a.step_size).is_equivalent(u.mm)

    def test_kernel_size(self, a: optika.metrology.AbstractSlopeErrorParameters):
        assert np.issubdtype(na.get_dtype(a.kernel_size), float)
        assert na.unit_normalized(a.kernel_size).is_equivalent(u.mm)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.metrology.SlopeErrorParameters(
            step_size=4 * u.mm,
            kernel_size=2 * u.mm,
        )
    ],
)
class TestSlopeErrorParameters(
    AbstractTestAbstractSlopeErrorParameters,
):
    pass
